########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import collections
import enum

import aioxmpp
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces

namespaces.xep0004_data = "jabber:x:data"


class Value(xso.XSO):
    TAG = (namespaces.xep0004_data, "value")

    value = xso.Text(default="")


class ValueElement(xso.AbstractElementType):
    def unpack(self, item):
        return item.value

    def pack(self, value):
        v = Value()
        v.value = value
        return v

    def get_xso_types(self):
        return [Value]


class Option(xso.XSO):
    TAG = (namespaces.xep0004_data, "option")

    label = xso.Attr(
        tag="label",
        default=None,
    )

    value = xso.ChildText(
        (namespaces.xep0004_data, "value"),
        default=None,
    )

    def validate(self):
        if self.value is None:
            raise ValueError("option is missing a value")


class OptionElement(xso.AbstractElementType):
    def unpack(self, item):
        return (item.value, item.label)

    def pack(self, value):
        value, label = value
        o = Option()
        o.value = value
        o.label = label
        return o

    def get_xso_types(self):
        return [Option]


class FieldType(enum.Enum):
    """
    Enumeration containing the field types defined in :xep:`4`.

    .. seealso::

       :attr:`Field.values`
          for important information regarding typing, restrictions, validation
          and constraints of values in that attribute.

    Each type has the following attributes and methods:

    .. automethod:: allow_upcast

    .. autoattribute:: has_options

    .. autoattribute:: is_multivalued

    Quotations in the following attribute descriptions are from said XEP.

    .. attribute:: BOOLEAN

       The ``"boolean"`` field:

          The field enables an entity to gather or provide an either-or choice
          between two options. The default value is "false".

       The :attr:`Field.values` sequence should contain zero or one elements.
       If it contains an element, it must be ``"0"``, ``"1"``, ``"false"``, or
       ``"true"``, in accordance with the XML Schema documents.

    .. attribute:: FIXED

       The ``"fixed"`` field:

          The field is intended for data description (e.g., human-readable text
          such as "section" headers) rather than data gathering or provision.
          The <value/> child SHOULD NOT contain newlines (the ``\\n`` and ``\\r``
          characters); instead an application SHOULD generate multiple fixed
          fields, each with one <value/> child.

       As such, the :attr:`Field.values` sequence should contain exactly one
       element. :attr:`Field.desc`, :attr:`Field.label`, :attr:`Field.options`
       and :attr:`Field.var` should be set to :data:`None` or empty containers.

    .. attribute:: HIDDEN

       The ``"hidden"`` field:

          The field is not shown to the form-submitting entity, but instead is
          returned with the form. The form-submitting entity SHOULD NOT modify
          the value of a hidden field, but MAY do so if such behavior is
          defined for the "using protocol".

       This type is commonly used for the ``var="FORM_TYPE"`` field, as
       specified in :xep:`68`.

    .. attribute:: JID_MULTI

       The ``"jid-multi"`` field:

          The field enables an entity to gather or provide multiple Jabber IDs.
          Each provided JID SHOULD be unique (as determined by comparison that
          includes application of the Nodeprep, Nameprep, and Resourceprep
          profiles of Stringprep as specified in XMPP Core), and duplicate JIDs
          MUST be ignored.

       As such, the :attr:`Field.values` sequence should contain zero or more
       strings representing Jabber IDs. :attr:`Field.options` should be empty.

    .. attribute:: JID_SINGLE

       The ``"jid-single"`` field:

          The field enables an entity to gather or provide a single Jabber ID.

       As such, the :attr:`Field.values` sequence should contain zero or one
       string representing a Jabber ID. :attr:`Field.options` should be empty.

    .. attribute:: LIST_MULTI

       The ``"list-multi"`` field:

          The field enables an entity to gather or provide one or more options
          from among many. A form-submitting entity chooses one or more items
          from among the options presented by the form-processing entity and
          MUST NOT insert new options. The form-submitting entity MUST NOT
          modify the order of items as received from the form-processing
          entity, since the order of items MAY be significant.

       Thus, :attr:`Field.values` should contain a subset of the keys of the
       :class:`Field.options` dictionary.

    .. attribute:: LIST_SINGLE

       The ``"list-single"`` field:

          The field enables an entity to gather or provide one option from
          among many. A form-submitting entity chooses one item from among the
          options presented by the form-processing entity and MUST NOT insert
          new options.

       Thus, :attr:`Field.values` should contain a zero or one of the keys of
       the :class:`Field.options` dictionary.

    .. attribute:: TEXT_MULTI

       The ``"text-multi"`` field:

          The field enables an entity to gather or provide multiple lines of
          text.

       Each string in the :attr:`Field.values` attribute should be a single
       line of text. Newlines are not allowed in data forms fields (due to the
       ambiguity between ``\\r`` and ``\\n`` and combinations thereof), which
       is why the text is split on the line endings.

    .. attribute:: TEXT_PRIVATE

       The ``"text-private"`` field:

          The field enables an entity to gather or provide a single line or
          word of text, which shall be obscured in an interface (e.g., with
          multiple instances of the asterisk character).

       The :attr:`Field.values` attribute should contain zero or one string
       without any newlines.

    .. attribute:: TEXT_SINGLE

       The ``"text-single"`` field:

          The field enables an entity to gather or provide a single line or
          word of text, which may be shown in an interface. This field type is
          the default and MUST be assumed if a form-submitting entity receives
          a field type it does not understand.

       The :attr:`Field.values` attribute should contain zero or one string
       without any newlines.

    """

    FIXED = "fixed"
    HIDDEN = "hidden"
    BOOLEAN = "boolean"
    TEXT_SINGLE = "text-single"
    TEXT_MULTI = "text-multi"
    TEXT_PRIVATE = "text-private"
    LIST_SINGLE = "list-single"
    LIST_MULTI = "list-multi"
    JID_SINGLE = "jid-single"
    JID_MULTI = "jid-multi"

    @property
    def has_options(self):
        """
        true for the ``list-`` field types, false otherwise.
        """
        return self.value.startswith("list-")

    @property
    def is_multivalued(self):
        """
        true for the ``-multi`` field types, false otherwise.
        """
        return self.value.endswith("-multi")

    def allow_upcast(self, to):
        """
        Return true if the field type may be upcast to the other field type
        `to`.

        This relation specifies when it is safe to transfer data from this
        field type to the given other field type `to`.

        This is the case if any of the following holds true:

        * `to` is equal to this type
        * this type is :attr:`TEXT_SINGLE` and `to` is :attr:`TEXT_PRIVATE`
        """

        if self == to:
            return True
        if self == FieldType.TEXT_SINGLE and to == FieldType.TEXT_PRIVATE:
            return True
        return False


class Field(xso.XSO):
    """
    Represent a single field in a Data Form.

    :param type_: Field type, must be one of the valid field types specified in
                  :xep:`4`.
    :type type_: :class:`FieldType`
    :param options: A mapping of values to labels defining the options in a
                    ``list-*`` field.
    :type options: :class:`dict` mapping :class:`str` to :class:`str`
    :param values: A sequence of values currently given for the field. Having
                   more than one value is only valid in ``*-multi`` fields.
    :type values: :class:`list` of :class:`str`
    :param desc: Description which can be shown in a tool-tip or similar,
                 without newlines.
    :type desc: :class:`str` or :data:`None`s
    :param label: Human-readable label to be shown next to the field input
    :type label: :class:`str` or :data:`None`
    :param required: Flag to indicate that the field is required
    :type required: :class:`bool`
    :param var: "ID" identifying the field uniquely inside the form. Only
                required for fields carrying a meaning (thus, not for
                ``fixed``).
    :type var: :class:`str` or :data:`None`

    The semantics of a :class:`Field` are different depending on where it
    occurs: in a :class:`Data`, it is a form field to be filled in, in a
    :class:`Item` it is a cell of a row and in a :class:`Reported` it
    represents a column header.

    .. attribute:: required

       A boolean flag indicating whether the field is required.

       If true, the XML serialisation will contain the corresponding
       ``<required/>`` tag.

    .. attribute:: desc

       Single line of description for the field. This attribute represents the
       ``<desc/>`` element from :xep:`4`.

    .. attribute:: values

       A sequence of strings representing the ``<value/>`` elements of the
       field, one string for each value.

       .. note::

          Since the requirements on the sequence of strings in :attr:`values`
          change depending on the :attr:`type_` attribute, validation and type
          conversion on assignment is very lax. The attribute accepts all
          sequences of strings, even if the field is for example a
          :attr:`FieldType.BOOLEAN` field, which allows for at most one string
          of a well-defined format (see the documentation there for the
          details).

          This makes it easy to inadvertendly generate invalid forms, which is
          why you should be using :class:`Form` subclasses when accessing forms
          from within normal code and some other, generic mechanism taking care
          of these details when showing forms in a UI framework to users. Note
          that devising such a mechanism is out of scope for :mod:`aioxmpp`, as
          every UI framework has different requirements.

    .. attribute:: options

       A dictionary mapping values to human-readable labels, representing the
       ``<option/>`` elements of the field.

    .. attribute:: var

       The uniquely identifying string of the (valued, that is,
       non-:attr:`FieldType.FIXED` field). Represents the ``var`` attribute of
       the field.

    .. attribute:: type_

       The type of the field. The :attr:`type_` must be a :class:`FieldType`
       enumeration value and determines restrictions and constraints on other
       attributes. See the :class:`FieldType` enumeration and :xep:`4` for
       details.

    .. attribute:: label

       The human-readable label for the field, representing the ``label``
       attribute of the field. May be :data:`None` if the label is omitted.

    """

    TAG = (namespaces.xep0004_data, "field")

    required = xso.ChildFlag(
        (namespaces.xep0004_data, "required"),
    )

    desc = xso.ChildText(
        (namespaces.xep0004_data, "desc"),
        default=None
    )

    values = xso.ChildValueList(
        type_=ValueElement()
    )

    options = xso.ChildValueMap(
        type_=OptionElement(),
        mapping_type=collections.OrderedDict,
    )

    var = xso.Attr(
        (None, "var"),
        default=None
    )

    type_ = xso.Attr(
        (None, "type"),
        type_=xso.EnumCDataType(
            FieldType,
        ),
        default=None,
    )

    label = xso.Attr(
        (None, "label"),
        default=None
    )

    def __init__(self, *,
                 type_=FieldType.TEXT_SINGLE,
                 options={},
                 values=[],
                 desc=None,
                 label=None,
                 required=False,
                 var=None):
        super().__init__()
        self.type_ = type_
        self.options.update(options)
        self.values[:] = values
        self.desc = desc
        self.label = label
        self.required = required
        self.var = var

    def validate(self):
        super().validate()

        if self.type_ != FieldType.FIXED and not self.var:
            raise ValueError("missing attribute var")

        if self.type_ is not None:
            if not self.type_.has_options and self.options:
                raise ValueError("unexpected option on non-list field")

            if not self.type_.is_multivalued and len(self.values) > 1:
                raise ValueError("too many values on non-multi field")

        values_list = [opt for opt in self.options.values()]
        values_set = set(values_list)

        if len(values_list) != len(values_set):
            raise ValueError("duplicate option value")


class AbstractItem(xso.XSO):
    fields = xso.ChildList([Field])


class Item(AbstractItem):
    """
    A single row in a report :class:`Data` object.

    .. attribute:: fields

       A sequence of :class:`Field` objects representing the cells of the row.
    """

    TAG = (namespaces.xep0004_data, "item")


class Reported(AbstractItem):
    """
    The table heading of a report :class:`Data` object.

    .. attribute:: fields

       A sequence of :class:`Field` objects representing the columns of the
       report or table.

    """

    TAG = (namespaces.xep0004_data, "reported")


class Instructions(xso.XSO):
    TAG = (namespaces.xep0004_data, "instructions")

    value = xso.Text(default="")


class InstructionsElement(xso.AbstractElementType):
    def unpack(self, item):
        return item.value

    def pack(self, value):
        v = Instructions()
        v.value = value
        return v

    def get_xso_types(self):
        return [Instructions]


class DataType(enum.Enum):
    """
    Enumeration containing the :class:`Data` types defined in :xep:`4`.

    Quotations in the following attribute descriptions are from :xep:`4`.

    .. attribute:: FORM

       The ``"form"`` type:

         The form-processing entity is asking the form-submitting entity to
         complete a form.

    .. attribute:: SUBMIT

       The ``"submit"`` type:

         The form-submitting entity is submitting data to the form-processing
         entity. The submission MAY include fields that were not provided in
         the empty form, but the form-processing entity MUST ignore any fields
         that it does not understand.

    .. attribute:: CANCEL

       The ``"cancel"`` type:

         The form-submitting entity has cancelled submission of data to the
         form-processing entity.

    .. attribute:: RESULT

       The ``"result"`` type:

         The form-processing entity is returning data (e.g., search results) to
         the form-submitting entity, or the data is a generic data set.
    """
    FORM = "form"
    SUBMIT = "submit"
    RESULT = "result"
    CANCEL = "cancel"


class Data(AbstractItem):
    """
    A :xep:`4` ``x`` element, that is, a Data Form.

    :param type_: Initial value for the :attr:`type_` attribute.

    .. attribute:: type_

       The ``type`` attribute of the form, represented by one of the members of
       the :class:`DataType` enumeration.

    .. attribute:: title

       The (optional) title of the form. Either a :class:`str` or :data:`None`.

    .. attribute:: instructions

       A sequence of strings which represent the instructions elements on the
       form.

    .. attribute:: fields

       If the :class:`Data` is a form, this is a sequence of :class:`Field`
       elements which represent the fields to be filled in.

       This does not make sense on :attr:`.DataType.RESULT` typed objects.

    .. attribute:: items

       If the :class:`Data` is a table, this is a sequence of :class:`Item`
       instances which represent the table rows.

       This only makes sense on :attr:`.DataType.RESULT` typed objects.

    .. attribute:: reported

       If the :class:`Data` is a table, this is a :class:`Reported` object
       representing the table header.

       This only makes sense on :attr:`.DataType.RESULT` typed objects.

    .. automethod:: get_form_type
    """

    TAG = (namespaces.xep0004_data, "x")

    type_ = xso.Attr(
        "type",
        type_=xso.EnumCDataType(DataType)
    )

    title = xso.ChildText(
        (namespaces.xep0004_data, "title"),
        default=None,
    )

    instructions = xso.ChildValueList(
        type_=InstructionsElement()
    )

    items = xso.ChildList([Item])

    reported = xso.Child([Reported], required=False)

    def __init__(self, type_):
        super().__init__()
        self.type_ = type_

    def _validate_result(self):
        if self.fields:
            raise ValueError("field in report result")

        fieldvars = {field.var for field in self.reported.fields}

        if not fieldvars:
            raise ValueError("empty report header")

        for item in self.items:
            itemvars = {field.var for field in item.fields}
            if itemvars != fieldvars:
                raise ValueError("field mismatch between row and header")

    def validate(self):
        super().validate()

        if (self.type_ != DataType.RESULT and
                (self.reported is not None or self.items)):
            raise ValueError("report in non-result")

        if (self.type_ == DataType.RESULT and
                (self.reported is not None or self.items)):
            self._validate_result()

    def get_form_type(self):
        """
        Extract the ``FORM_TYPE`` from the fields.

        :return: ``FORM_TYPE`` value or :data:`None`
        :rtype: :class:`str` or :data:`None`

        Return :data:`None` if no well-formed ``FORM_TYPE`` field is found in
        the list of fields.

        .. versionadded:: 0.8
        """

        for field in self.fields:
            if field.var == "FORM_TYPE" and field.type_ == FieldType.HIDDEN:
                if len(field.values) != 1:
                    return None
                return field.values[0]


aioxmpp.Message.xep0004_data = xso.ChildList([Data])



########################################################################
# File name: test_e2e.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp
import aioxmpp.stream
import aioxmpp.xso

from aioxmpp.testutils import get_timeout

from aioxmpp.e2etest import (
    blocking_timed,
    TestCase,
)


@aioxmpp.IQ.as_payload_class
class MadeUpIQPayload(aioxmpp.xso.XSO):
    TAG = "made-up", "made-up"


class TestConnect(TestCase):
    @blocking_timed
    @asyncio.coroutine
    def test_provisioner_works_in_general(self):
        connected = yield from self.provisioner.get_connected_client()
        self.assertTrue(connected.running)
        self.assertTrue(connected.established)


class TestMessaging(TestCase):
    @blocking_timed
    @asyncio.coroutine
    def test_message_from_a_to_b(self):
        a = yield from self.provisioner.get_connected_client()
        b = yield from self.provisioner.get_connected_client()

        fut = asyncio.Future()

        def cb(msg):
            fut.set_result(msg)

        with aioxmpp.stream.message_handler(
                b.stream,
                aioxmpp.MessageType.CHAT,
                a.local_jid,
                cb):

            msg_sent = aioxmpp.Message(
                to=b.local_jid,
                type_=aioxmpp.MessageType.CHAT,

            )
            msg_sent.body[aioxmpp.structs.LanguageTag.fromstr("en")] = \
                "Hello World!"

            yield from a.send(msg_sent)

            msg_rcvd = yield from fut

        self.assertEqual(
            msg_rcvd.body[aioxmpp.structs.LanguageTag.fromstr("en")],
            "Hello World!"
        )


class TestMisc(TestCase):
    @blocking_timed
    @asyncio.coroutine
    def test_receive_response_from_iq_to_bare_explicit_self(self):
        c = yield from self.provisioner.get_connected_client()

        iq = aioxmpp.IQ(
            to=c.local_jid.bare(),
            type_=aioxmpp.IQType.GET,
            payload=MadeUpIQPayload()
        )

        with self.assertRaises(aioxmpp.errors.XMPPCancelError):
            yield from c.send(iq)

    @blocking_timed
    @asyncio.coroutine
    def test_receive_response_from_iq_to_bare_self_using_None(self):
        c = yield from self.provisioner.get_connected_client()

        iq = aioxmpp.IQ(
            to=None,
            type_=aioxmpp.IQType.GET,
            payload=MadeUpIQPayload()
        )

        with self.assertRaises(aioxmpp.errors.XMPPCancelError):
            yield from c.send(iq)

    @blocking_timed
    @asyncio.coroutine
    def test_handle_malformed_iq_error_gracefully(self):
        c = yield from self.provisioner.get_connected_client()

        @asyncio.coroutine
        def handler(iq):
            # this is awful, but does the trick
            c.stream._xmlstream.data_received(
                "<iq type='error' from='{}' to='{}' id='{}'/>".format(
                    c.local_jid,
                    c.local_jid,
                    iq.id_,
                )
            )

        c.stream.register_iq_request_handler(
            aioxmpp.IQType.GET,
            MadeUpIQPayload,
            handler,
        )

        iq = aioxmpp.IQ(
            to=c.local_jid,
            type_=aioxmpp.IQType.GET,
            payload=MadeUpIQPayload()
        )

        with self.assertRaises(aioxmpp.errors.ErroneousStanza):
            yield from c.stream.send(iq)

    @blocking_timed
    @asyncio.coroutine
    def test_handle_id_less_IQ_request_gracefully(self):
        c = yield from self.provisioner.get_connected_client()

        # let’s get even more brutal here
        c.stream._xmlstream.data_received(
            "<iq type='get' to='{}'/>".format(c.local_jid).encode('utf-8')
        )

        # now we send a simple IQ which shows us that the stream survived this
        # attack

        iq = aioxmpp.IQ(
            to=c.local_jid.bare(),
            type_=aioxmpp.IQType.GET,
            payload=MadeUpIQPayload()
        )

        with self.assertRaises(aioxmpp.errors.XMPPCancelError):
            yield from c.send(iq)

    @blocking_timed
    @asyncio.coroutine
    def test_exception_from_non_wellformed(self):
        c = yield from self.provisioner.get_connected_client()

        msg = aioxmpp.Message(
            to=c.local_jid,
            type_=aioxmpp.MessageType.NORMAL,
        )
        msg.body[None] = "foo\u0000"

        with self.assertRaisesRegex(ValueError, "not allowed"):
            yield from c.send(msg)

        msg.body[None] = "foo"

        yield from c.send(msg)



########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import contextlib
import unittest
import sys

import aioxmpp.service as service
import aioxmpp.disco.service as disco_service
import aioxmpp.disco.xso as disco_xso
import aioxmpp.stanza as stanza
import aioxmpp.structs as structs
import aioxmpp.errors as errors

from aioxmpp.utils import namespaces

from aioxmpp.testutils import (
    make_connected_client,
    run_coroutine,
    CoroutineMock,
)


TEST_JID = structs.JID.fromstr("foo@bar.example")


class TestNode(unittest.TestCase):
    def test_init(self):
        n = disco_service.Node()
        self.assertSequenceEqual(
            [],
            list(n.iter_identities(unittest.mock.sentinel.stanza))
        )
        self.assertSetEqual(
            {namespaces.xep0030_info},
            set(n.iter_features(unittest.mock.sentinel.stanza))
        )
        self.assertSequenceEqual(
            [],
            list(n.iter_items(unittest.mock.sentinel.stanza))
        )

    def test_register_feature_adds_the_feature(self):
        n = disco_service.Node()
        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        n.register_feature("uri:foo")

        self.assertSetEqual(
            {
                "uri:foo",
                namespaces.xep0030_info
            },
            set(n.iter_features(unittest.mock.sentinel.stanza))
        )

        cb.assert_called_with()

    def test_iter_features_works_without_argument(self):
        n = disco_service.Node()
        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        n.register_feature("uri:foo")

        self.assertSetEqual(
            {
                "uri:foo",
                namespaces.xep0030_info
            },
            set(n.iter_features())
        )

        cb.assert_called_with()

    def test_register_feature_prohibits_duplicate_registration(self):
        n = disco_service.Node()
        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        n.register_feature("uri:bar")
        cb.mock_calls.clear()

        with self.assertRaisesRegex(ValueError,
                                    "feature already claimed"):
            n.register_feature("uri:bar")

        self.assertSetEqual(
            {
                "uri:bar",
                namespaces.xep0030_info
            },
            set(n.iter_features(unittest.mock.sentinel.stanza))
        )

        self.assertFalse(cb.mock_calls)

    def test_register_feature_prohibits_registration_of_xep0030_features(self):
        n = disco_service.Node()

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        with self.assertRaisesRegex(ValueError,
                                    "feature already claimed"):
            n.register_feature(namespaces.xep0030_info)

        self.assertFalse(cb.mock_calls)

    def test_unregister_feature_removes_the_feature(self):
        n = disco_service.Node()
        n.register_feature("uri:foo")
        n.register_feature("uri:bar")

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        self.assertSetEqual(
            {
                "uri:foo",
                "uri:bar",
                namespaces.xep0030_info
            },
            set(n.iter_features(unittest.mock.sentinel.stanza))
        )

        n.unregister_feature("uri:foo")

        cb.assert_called_with()
        cb.mock_calls.clear()

        self.assertSetEqual(
            {
                "uri:bar",
                namespaces.xep0030_info
            },
            set(n.iter_features(unittest.mock.sentinel.stanza))
        )

        n.unregister_feature("uri:bar")

        self.assertSetEqual(
            {
                namespaces.xep0030_info
            },
            set(n.iter_features(unittest.mock.sentinel.stanza))
        )

        cb.assert_called_with()
        cb.mock_calls.clear()

    def test_unregister_feature_prohibts_removal_of_nonexistant_feature(self):
        n = disco_service.Node()

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        with self.assertRaises(KeyError):
            n.unregister_feature("uri:foo")

        self.assertFalse(cb.mock_calls)

    def test_unregister_feature_prohibts_removal_of_xep0030_features(self):
        n = disco_service.Node()

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        with self.assertRaises(KeyError):
            n.unregister_feature(namespaces.xep0030_info)

        self.assertSetEqual(
            {
                namespaces.xep0030_info
            },
            set(n.iter_features(unittest.mock.sentinel.stanza))
        )

        self.assertFalse(cb.mock_calls)

    def test_register_identity_defines_identity(self):
        n = disco_service.Node()

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        n.register_identity(
            "client", "pc"
        )

        self.assertSetEqual(
            {
                ("client", "pc", None, None),
            },
            set(n.iter_identities(unittest.mock.sentinel.stanza))
        )

        cb.assert_called_with()

    def test_iter_identities_works_without_stanza(self):
        n = disco_service.Node()

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        n.register_identity(
            "client", "pc"
        )

        self.assertSetEqual(
            {
                ("client", "pc", None, None),
            },
            set(n.iter_identities())
        )

        cb.assert_called_with()

    def test_register_identity_prohibits_duplicate_registration(self):
        n = disco_service.Node()

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        n.register_identity(
            "client", "pc"
        )

        cb.assert_called_with()
        cb.mock_calls.clear()

        with self.assertRaisesRegex(ValueError,
                                    "identity already claimed"):
            n.register_identity("client", "pc")

        self.assertFalse(cb.mock_calls)

        self.assertSetEqual(
            {
                ("client", "pc", None, None),
            },
            set(n.iter_identities(unittest.mock.sentinel.stanza))
        )

    def test_register_identity_with_names(self):
        n = disco_service.Node()

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        n.register_identity(
            "client", "pc",
            names={
                structs.LanguageTag.fromstr("en"): "test identity",
                structs.LanguageTag.fromstr("de"): "Testidentität",
            }
        )

        cb.assert_called_with()

        self.assertSetEqual(
            {
                ("client", "pc",
                 structs.LanguageTag.fromstr("en"), "test identity"),
                ("client", "pc",
                 structs.LanguageTag.fromstr("de"), "Testidentität"),
            },
            set(n.iter_identities(unittest.mock.sentinel.stanza))
        )

    def test_set_identity_names_updates_text_and_emits_cb(self):
        n = disco_service.Node()

        n.register_identity(
            "client", "pc",
            names={
                structs.LanguageTag.fromstr("en"): "default identity",
                structs.LanguageTag.fromstr("de"): "Standardidentität",
            }
        )

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        n.set_identity_names(
            "client", "pc",
            names={
                structs.LanguageTag.fromstr("en"): "test identity",
                structs.LanguageTag.fromstr("de"): "Testidentität",
            }
        )

        cb.assert_called_with()

        self.assertSetEqual(
            {
                ("client", "pc",
                 structs.LanguageTag.fromstr("en"), "test identity"),
                ("client", "pc",
                 structs.LanguageTag.fromstr("de"), "Testidentität"),
            },
            set(n.iter_identities(unittest.mock.sentinel.stanza))
        )

    def test_set_identity_names_rejects_if_identity_not_registered(self):
        n = disco_service.Node()

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        with self.assertRaisesRegex(
                ValueError,
                r"identity not registered"):

            n.set_identity_names(
                "client", "pc",
                names={
                    structs.LanguageTag.fromstr("en"): "test identity",
                    structs.LanguageTag.fromstr("de"): "Testidentität",
                }
            )

        cb.assert_not_called()

    def test_unregister_identity_prohibits_removal_of_last_identity(self):
        n = disco_service.Node()

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        n.register_identity(
            "client", "pc",
            names={
                structs.LanguageTag.fromstr("en"): "test identity",
                structs.LanguageTag.fromstr("de"): "Testidentität",
            }
        )

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        with self.assertRaisesRegex(ValueError,
                                    "cannot remove last identity"):
            n.unregister_identity(
                "client", "pc",
            )

        self.assertFalse(cb.mock_calls)

    def test_unregister_identity_prohibits_removal_of_undeclared_identity(
            self):
        n = disco_service.Node()

        n.register_identity(
            "client", "pc",
            names={
                structs.LanguageTag.fromstr("en"): "test identity",
                structs.LanguageTag.fromstr("de"): "Testidentität",
            }
        )

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        with self.assertRaises(KeyError):
            n.unregister_identity("foo", "bar")

        self.assertFalse(cb.mock_calls)

    def test_unregister_identity_removes_identity(self):
        n = disco_service.Node()

        n.register_identity(
            "client", "pc",
            names={
                structs.LanguageTag.fromstr("en"): "test identity",
                structs.LanguageTag.fromstr("de"): "Testidentität",
            }
        )

        n.register_identity(
            "foo", "bar"
        )

        self.assertSetEqual(
            {
                ("client", "pc",
                 structs.LanguageTag.fromstr("en"), "test identity"),
                ("client", "pc",
                 structs.LanguageTag.fromstr("de"), "Testidentität"),
                ("foo", "bar", None, None),
            },
            set(n.iter_identities(unittest.mock.sentinel.stanza))
        )

        cb = unittest.mock.Mock()
        n.on_info_changed.connect(cb)

        n.unregister_identity("foo", "bar")

        cb.assert_called_with()

        self.assertSetEqual(
            {
                ("client", "pc",
                 structs.LanguageTag.fromstr("en"), "test identity"),
                ("client", "pc",
                 structs.LanguageTag.fromstr("de"), "Testidentität"),
            },
            set(n.iter_identities(unittest.mock.sentinel.stanza))
        )

    def test_iter_items_works_without_argument(self):
        n = disco_service.Node()
        self.assertSequenceEqual(
            list(n.iter_items()),
            []
        )

    def test_as_info_xso(self):
        n = disco_service.Node()

        features = [
            "http://jabber.org/protocol/disco#info",
            unittest.mock.sentinel.f1,
            unittest.mock.sentinel.f2,
            unittest.mock.sentinel.f3,
        ]

        identities = [
            ("cat1", "t1",
             structs.LanguageTag.fromstr("lang-a"), "name11"),
            ("cat1", "t1",
             structs.LanguageTag.fromstr("lang-b"), "name12"),
            ("cat2", "t2", None, "name2"),
            ("cat3", "t3", None, None),
        ]

        with contextlib.ExitStack() as stack:
            iter_features = stack.enter_context(
                unittest.mock.patch.object(n, "iter_features")
            )
            iter_features.return_value = iter(features)

            iter_identities = stack.enter_context(
                unittest.mock.patch.object(n, "iter_identities")
            )
            iter_identities.return_value = iter(identities)

            iter_items = stack.enter_context(
                unittest.mock.patch.object(n, "iter_items")
            )

            result = n.as_info_xso()

        self.assertIsInstance(
            result,
            disco_xso.InfoQuery,
        )

        iter_items.assert_not_called()

        iter_features.assert_called_once_with(None)
        iter_identities.assert_called_once_with(None)

        self.assertSetEqual(
            result.features,
            set(features),
        )

        self.assertCountEqual(
            [
                (i.category, i.type_, i.lang, i.name)
                for i in result.identities
            ],
            identities,
        )

    def test_as_info_xso_with_stanza(self):
        n = disco_service.Node()

        features = [
            "http://jabber.org/protocol/disco#info",
            unittest.mock.sentinel.f1,
        ]

        identities = [
            ("cat1", "t1",
             structs.LanguageTag.fromstr("lang-a"), "name11"),
            ("cat1", "t1",
             structs.LanguageTag.fromstr("lang-b"), "name12"),
        ]

        with contextlib.ExitStack() as stack:
            iter_features = stack.enter_context(
                unittest.mock.patch.object(n, "iter_features")
            )
            iter_features.return_value = iter(features)

            iter_identities = stack.enter_context(
                unittest.mock.patch.object(n, "iter_identities")
            )
            iter_identities.return_value = iter(identities)

            iter_items = stack.enter_context(
                unittest.mock.patch.object(n, "iter_items")
            )

            result = n.as_info_xso(unittest.mock.sentinel.stanza)

        self.assertIsInstance(
            result,
            disco_xso.InfoQuery,
        )

        iter_items.assert_not_called()

        iter_features.assert_called_once_with(unittest.mock.sentinel.stanza)
        iter_identities.assert_called_once_with(unittest.mock.sentinel.stanza)

        self.assertSetEqual(
            result.features,
            set(features),
        )

        self.assertCountEqual(
            [
                (i.category, i.type_, i.lang, i.name)
                for i in result.identities
            ],
            identities,
        )


class TestStaticNode(unittest.TestCase):
    def setUp(self):
        self.n = disco_service.StaticNode()

    def test_is_Node(self):
        self.assertIsInstance(self.n, disco_service.Node)

    def test_add_items(self):
        item1 = disco_xso.Item(TEST_JID.replace(localpart="foo"))
        item2 = disco_xso.Item(TEST_JID.replace(localpart="bar"))
        self.n.items.append(item1)
        self.n.items.append(item2)

        self.assertSequenceEqual(
            [
                item1,
                item2
            ],
            list(self.n.iter_items(unittest.mock.sentinel.jid))
        )

    def test_iter_items_works_without_argument(self):
        self.assertSequenceEqual(
            list(self.n.iter_items()),
            []
        )

    def test_clone(self):
        other_node = unittest.mock.Mock([
            "iter_features",
            "iter_identities",
            "iter_items",
        ])

        features = [
            "http://jabber.org/protocol/disco#info",
            unittest.mock.sentinel.f1,
            unittest.mock.sentinel.f2,
            unittest.mock.sentinel.f3,
        ]

        identities = [
            (unittest.mock.sentinel.cat1, unittest.mock.sentinel.t1,
             unittest.mock.sentinel.lang11, unittest.mock.sentinel.name11),
            (unittest.mock.sentinel.cat1, unittest.mock.sentinel.t1,
             unittest.mock.sentinel.lang12, unittest.mock.sentinel.name12),
            (unittest.mock.sentinel.cat2, unittest.mock.sentinel.t2,
             None, unittest.mock.sentinel.name2),
            (unittest.mock.sentinel.cat3, unittest.mock.sentinel.t3,
             None, None),
        ]

        items = [
            unittest.mock.sentinel.item1,
            unittest.mock.sentinel.item2,
        ]

        other_node.iter_features.return_value = iter(features)
        other_node.iter_identities.return_value = iter(identities)
        other_node.iter_items.return_value = iter(items)

        n = disco_service.StaticNode.clone(other_node)

        self.assertIsInstance(n, disco_service.StaticNode)

        other_node.iter_features.assert_called_once_with()
        other_node.iter_identities.assert_called_once_with()
        other_node.iter_items.assert_called_once_with()

        self.assertCountEqual(
            features,
            list(n.iter_features()),
        )

        self.assertCountEqual(
            identities,
            list(n.iter_identities()),
        )

        self.assertCountEqual(
            items,
            list(n.iter_items()),
        )


class TestDiscoServer(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.s = disco_service.DiscoServer(self.cc)
        self.cc.reset_mock()

        self.request_iq = stanza.IQ(
            structs.IQType.GET,
            from_=structs.JID.fromstr("user@foo.example/res1"),
            to=structs.JID.fromstr("user@bar.example/res2"))
        self.request_iq.autoset_id()
        self.request_iq.payload = disco_xso.InfoQuery()

        self.request_items_iq = stanza.IQ(
            structs.IQType.GET,
            from_=structs.JID.fromstr("user@foo.example/res1"),
            to=structs.JID.fromstr("user@bar.example/res2"))
        self.request_items_iq.autoset_id()
        self.request_items_iq.payload = disco_xso.ItemsQuery()

    def test_is_Service_subclass(self):
        self.assertTrue(issubclass(
            disco_service.DiscoServer,
            service.Service))

    def test_setup(self):
        cc = make_connected_client()
        s = disco_service.DiscoServer(cc)

        self.assertCountEqual(
            [
                unittest.mock.call.stream.register_iq_request_handler(
                    structs.IQType.GET,
                    disco_xso.InfoQuery,
                    s.handle_info_request
                ),
                unittest.mock.call.stream.register_iq_request_handler(
                    structs.IQType.GET,
                    disco_xso.ItemsQuery,
                    s.handle_items_request
                )
            ],
            cc.mock_calls
        )

    def test_shutdown(self):
        run_coroutine(self.s.shutdown())
        self.assertCountEqual(
            [
                unittest.mock.call.stream.unregister_iq_request_handler(
                    structs.IQType.GET,
                    disco_xso.InfoQuery
                ),
                unittest.mock.call.stream.unregister_iq_request_handler(
                    structs.IQType.GET,
                    disco_xso.ItemsQuery
                ),
            ],
            self.cc.mock_calls
        )

    def test_handle_info_request_is_decorated(self):
        self.assertTrue(
            service.is_iq_handler(
                structs.IQType.GET,
                disco_xso.InfoQuery,
                disco_service.DiscoServer.handle_info_request,
            )
        )

    def test_handle_items_request_is_decorated(self):
        self.assertTrue(
            service.is_iq_handler(
                structs.IQType.GET,
                disco_xso.ItemsQuery,
                disco_service.DiscoServer.handle_items_request,
            )
        )

    def test_default_response(self):
        response = run_coroutine(self.s.handle_info_request(self.request_iq))

        self.assertSetEqual(
            {namespaces.xep0030_info},
            response.features,
        )

        self.assertSetEqual(
            {
                ("client", "bot",
                 "aioxmpp default identity",
                 structs.LanguageTag.fromstr("en")),
            },
            set((item.category, item.type_,
                 item.name, item.lang) for item in response.identities)
        )

        self.assertFalse(response.node)

    def test_nonexistant_node_response(self):
        self.request_iq.payload.node = "foobar"
        with self.assertRaises(errors.XMPPModifyError) as ctx:
            run_coroutine(self.s.handle_info_request(self.request_iq))

        self.assertEqual(
            errors.ErrorCondition.ITEM_NOT_FOUND,
            ctx.exception.condition
        )

    def test_register_feature_produces_it_in_response(self):
        self.s.register_feature("uri:foo")
        self.s.register_feature("uri:bar")

        response = run_coroutine(self.s.handle_info_request(self.request_iq))

        self.assertSetEqual(
            {"uri:foo", "uri:bar", namespaces.xep0030_info},
            response.features,
        )

    def test_unregister_feature_removes_it_from_response(self):
        self.s.register_feature("uri:foo")
        self.s.register_feature("uri:bar")

        self.s.unregister_feature("uri:bar")

        response = run_coroutine(self.s.handle_info_request(self.request_iq))

        self.assertSetEqual(
            {"uri:foo", namespaces.xep0030_info},
            response.features
        )

    def test_unregister_feature_raises_KeyError_if_feature_has_not_been_registered(self):
        with self.assertRaisesRegex(KeyError, "uri:foo"):
            self.s.unregister_feature("uri:foo")

    def test_unregister_feature_disallows_unregistering_disco_info_feature(self):
        with self.assertRaises(KeyError):
            self.s.unregister_feature(namespaces.xep0030_info)

    def test_register_identity_produces_it_in_response(self):
        self.s.register_identity(
            "client", "pc"
        )
        self.s.register_identity(
            "hierarchy", "branch"
        )

        response = run_coroutine(self.s.handle_info_request(self.request_iq))

        self.assertSetEqual(
            {
                ("client", "pc", None, None),
                ("hierarchy", "branch", None, None),
                ("client", "bot", "aioxmpp default identity",
                 structs.LanguageTag.fromstr("en")),
            },
            set((item.category, item.type_,
                 item.name, item.lang) for item in response.identities)
        )

    def test_unregister_identity_removes_it_from_response(self):
        self.s.register_identity(
            "client", "pc"
        )

        self.s.unregister_identity("client", "bot")

        self.s.register_identity(
            "hierarchy", "branch"
        )

        self.s.unregister_identity("hierarchy", "branch")

        response = run_coroutine(self.s.handle_info_request(self.request_iq))

        self.assertSetEqual(
            {
                ("client", "pc", None, None),
            },
            set((item.category, item.type_,
                 item.name, item.lang) for item in response.identities)
        )

    def test_unregister_identity_raises_KeyError_if_not_registered(self):
        with self.assertRaisesRegex(KeyError, r"\('client', 'pc'\)"):
            self.s.unregister_identity("client", "pc")

    def test_register_identity_with_names(self):
        self.s.register_identity(
            "client", "pc",
            names={
                structs.LanguageTag.fromstr("en"): "test identity",
                structs.LanguageTag.fromstr("de"): "Testidentität",
            }
        )

        self.s.unregister_identity("client", "bot")

        response = run_coroutine(self.s.handle_info_request(self.request_iq))

        self.assertSetEqual(
            {
                ("client", "pc",
                 "test identity",
                 structs.LanguageTag.fromstr("en")),
                ("client", "pc",
                 "Testidentität",
                 structs.LanguageTag.fromstr("de")),
            },
            set((item.category, item.type_,
                 item.name, item.lang) for item in response.identities)
        )

    def test_register_identity_disallows_duplicates(self):
        self.s.register_identity("client", "pc")
        with self.assertRaisesRegex(ValueError, "identity already claimed"):
            self.s.register_identity("client", "pc")

    def test_register_feature_disallows_duplicates(self):
        self.s.register_feature("uri:foo")
        with self.assertRaisesRegex(ValueError, "feature already claimed"):
            self.s.register_feature("uri:foo")
        with self.assertRaisesRegex(ValueError, "feature already claimed"):
            self.s.register_feature(namespaces.xep0030_info)

    def test_mount_node_produces_response(self):
        node = disco_service.StaticNode()
        node.register_identity("hierarchy", "leaf")

        self.s.mount_node("foo", node)

        self.request_iq.payload.node = "foo"
        response = run_coroutine(self.s.handle_info_request(self.request_iq))

        self.assertSetEqual(
            {
                ("hierarchy", "leaf", None, None),
            },
            set((item.category, item.type_,
                 item.name, item.lang) for item in response.identities)
        )

    def test_mount_node_without_identity_produces_item_not_found(self):
        node = disco_service.StaticNode()

        self.s.mount_node("foo", node)

        self.request_iq.payload.node = "foo"
        with self.assertRaises(errors.XMPPModifyError):
            run_coroutine(self.s.handle_info_request(self.request_iq))

    def test_unmount_node(self):
        node = disco_service.StaticNode()
        node.register_identity("hierarchy", "leaf")

        self.s.mount_node("foo", node)

        self.s.unmount_node("foo")

        self.request_iq.payload.node = "foo"
        with self.assertRaises(errors.XMPPModifyError):
            run_coroutine(self.s.handle_info_request(self.request_iq))

    def test_default_items_response(self):
        response = run_coroutine(
            self.s.handle_items_request(self.request_items_iq)
        )
        self.assertIsInstance(response, disco_xso.ItemsQuery)
        self.assertSequenceEqual(
            [],
            response.items
        )

    def test_items_query_returns_item_not_found_for_unknown_node(self):
        self.request_items_iq.payload.node = "foobar"
        with self.assertRaises(errors.XMPPModifyError):
            run_coroutine(
                self.s.handle_items_request(self.request_items_iq)
            )

    def test_items_query_returns_items_of_mounted_node(self):
        item1 = disco_xso.Item(TEST_JID.replace(localpart="foo"))
        item2 = disco_xso.Item(TEST_JID.replace(localpart="bar"))

        node = disco_service.StaticNode()
        node.register_identity("hierarchy", "leaf")
        node.items.append(item1)
        node.items.append(item2)

        self.s.mount_node("foo", node)

        self.request_items_iq.payload.node = "foo"
        response = run_coroutine(
            self.s.handle_items_request(self.request_items_iq)
        )

        self.assertSequenceEqual(
            [item1, item2],
            response.items
        )

    def test_items_query_forwards_stanza(self):
        node = unittest.mock.Mock()
        node.iter_items.return_value = iter([])

        self.s.mount_node("foo", node)

        self.request_items_iq.payload.node = "foo"
        run_coroutine(
            self.s.handle_items_request(self.request_items_iq)
        )

        node.iter_items.assert_called_once_with(
            self.request_items_iq
        )

    def test_info_query_forwards_stanza(self):
        node = unittest.mock.Mock()

        self.s.mount_node("foo", node)
        self.request_iq.payload.node = "foo"

        result = run_coroutine(
            self.s.handle_info_request(self.request_iq)
        )

        node.as_info_xso.assert_called_once_with(
            self.request_iq
        )

        self.assertEqual(result, node.as_info_xso())

    def test_info_query_sets_node(self):
        node = unittest.mock.Mock()

        self.s.mount_node("foo", node)
        self.request_iq.payload.node = "foo"

        result = run_coroutine(
            self.s.handle_info_request(self.request_iq)
        )

        node.as_info_xso.assert_called_once_with(
            self.request_iq
        )

        self.assertEqual(result.node, "foo")
        self.assertEqual(result, node.as_info_xso())


class TestDiscoClient(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.s = disco_service.DiscoClient(self.cc)
        self.cc.reset_mock()

        self.request_iq = stanza.IQ(
            structs.IQType.GET,
            from_=structs.JID.fromstr("user@foo.example/res1"),
            to=structs.JID.fromstr("user@bar.example/res2"))
        self.request_iq.autoset_id()
        self.request_iq.payload = disco_xso.InfoQuery()

        self.request_items_iq = stanza.IQ(
            structs.IQType.GET,
            from_=structs.JID.fromstr("user@foo.example/res1"),
            to=structs.JID.fromstr("user@bar.example/res2"))
        self.request_items_iq.autoset_id()
        self.request_items_iq.payload = disco_xso.ItemsQuery()

    def test_is_Service_subclass(self):
        self.assertTrue(issubclass(
            disco_service.DiscoClient,
            service.Service))

    def test_send_and_decode_info_query(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        node = "foobar"
        response = disco_xso.InfoQuery()

        self.cc.send.return_value = response

        result = run_coroutine(
            self.s.send_and_decode_info_query(to, node)
        )

        self.assertEqual(result, response)

        self.assertEqual(
            1,
            len(self.cc.send.mock_calls)
        )

        call, = self.cc.send.mock_calls
        # call[1] are args
        request_iq, = call[1]

        self.assertEqual(
            to,
            request_iq.to
        )
        self.assertEqual(
            structs.IQType.GET,
            request_iq.type_
        )
        self.assertIsInstance(request_iq.payload, disco_xso.InfoQuery)
        self.assertFalse(request_iq.payload.features)
        self.assertFalse(request_iq.payload.identities)
        self.assertIs(request_iq.payload.node, node)

    def test_uses_LRUDict(self):
        with contextlib.ExitStack() as stack:
            LRUDict = stack.enter_context(unittest.mock.patch(
                "aioxmpp.cache.LRUDict",
                new=unittest.mock.MagicMock()
            ))

            self.s = disco_service.DiscoClient(self.cc)

        self.assertCountEqual(
            LRUDict.mock_calls,
            [
                unittest.mock.call(),
                unittest.mock.call(),
            ]
        )

        LRUDict().__getitem__.side_effect = KeyError

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:

            run_coroutine(
                self.s.query_info(unittest.mock.sentinel.jid,
                                  node=unittest.mock.sentinel.node)
            )

        LRUDict().__setitem__.assert_called_once_with(
            (unittest.mock.sentinel.jid, unittest.mock.sentinel.node),
            unittest.mock.ANY,
        )
        LRUDict().__setitem__.reset_mock()

        run_coroutine(
            self.s.query_items(TEST_JID,
                               node="some node")
        )

        LRUDict().__setitem__.assert_called_once_with(
            (TEST_JID, "some node"),
            unittest.mock.ANY,
        )

    def test_info_cache_size(self):
        self.assertEqual(
            self.s.info_cache_size,
            10000,
        )

        self.assertEqual(
            self.s.info_cache_size,
            self.s._info_pending.maxsize,
        )

    def test_info_cache_size_is_settable(self):
        self.s.info_cache_size = 5

        self.assertEqual(
            self.s.info_cache_size,
            5,
        )

        self.assertEqual(
            self.s._info_pending.maxsize,
            self.s.info_cache_size,
        )

    def test_items_cache_size(self):
        self.assertEqual(
            self.s.items_cache_size,
            100,
        )

        self.assertEqual(
            self.s.items_cache_size,
            self.s._items_pending.maxsize,
        )

    def test_items_cache_size_is_settable(self):
        self.s.items_cache_size = 5

        self.assertEqual(
            self.s.items_cache_size,
            5,
        )

        self.assertEqual(
            self.s._items_pending.maxsize,
            self.s.items_cache_size,
        )

    def test_query_info(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = {}

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            send_and_decode.return_value = response

            result = run_coroutine(
                self.s.query_info(to)
            )

        send_and_decode.assert_called_with(to, None)
        self.assertIs(response, result)

    def test_query_response_leads_to_signal_emission(self):
        handler = unittest.mock.Mock()
        handler.return_value = None

        to = structs.JID.fromstr("user@foo.example/res1")
        response = {}

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            send_and_decode.return_value = response

            self.s.on_info_result.connect(handler)

            run_coroutine(
                self.s.query_info(to)
            )

        handler.assert_called_with(to, None, response)

    def test_query_response_for_node_leads_to_signal_emission(self):
        handler = unittest.mock.Mock()
        handler.return_value = None

        to = structs.JID.fromstr("user@foo.example/res1")
        response = {}

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            send_and_decode.return_value = response

            self.s.on_info_result.connect(handler)

            run_coroutine(
                self.s.query_info(to, node="foo")
            )

        handler.assert_called_with(to, "foo", response)

    def test_query_info_with_node(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = {}

        with self.assertRaises(TypeError):
            self.s.query_info(to, "foobar")

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            send_and_decode.return_value = response

            result = run_coroutine(
                self.s.query_info(to, node="foobar")
            )

        send_and_decode.assert_called_with(to, "foobar")
        self.assertIs(result, response)

    def test_query_info_caches(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = {}

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            send_and_decode.return_value = response

            result1 = run_coroutine(
                self.s.query_info(to, node="foobar")
            )
            result2 = run_coroutine(
                self.s.query_info(to, node="foobar")
            )

            self.assertIs(result1, response)
            self.assertIs(result2, response)

        self.assertEqual(
            1,
            len(send_and_decode.mock_calls)
        )

    def test_query_info_does_not_cache_if_no_cache_is_false(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = {}

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            send_and_decode.return_value = response

            result1 = run_coroutine(
                self.s.query_info(to, node="foobar", no_cache=True)
            )
            result2 = run_coroutine(
                self.s.query_info(to, node="foobar")
            )

            self.assertIs(result1, response)
            self.assertIs(result2, response)

        self.assertEqual(
            2,
            len(send_and_decode.mock_calls)
        )

    def test_query_info_with_no_cache_uses_cached_result(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = {}

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            send_and_decode.return_value = response

            result1 = run_coroutine(
                self.s.query_info(to, node="foobar")
            )
            result2 = run_coroutine(
                self.s.query_info(to, node="foobar", no_cache=True)
            )

            self.assertIs(result1, response)
            self.assertIs(result2, response)

        self.assertEqual(
            1,
            len(send_and_decode.mock_calls)
        )

    def test_query_info_reraises_and_aliases_exception(self):
        to = structs.JID.fromstr("user@foo.example/res1")

        ncall = 0

        @asyncio.coroutine
        def mock(*args, **kwargs):
            nonlocal ncall
            ncall += 1
            if ncall == 1:
                raise errors.XMPPCancelError(
                    condition=errors.ErrorCondition.FEATURE_NOT_IMPLEMENTED
                )
            else:
                raise ConnectionError()

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=mock):

            task1 = asyncio.ensure_future(
                self.s.query_info(to, node="foobar")
            )
            task2 = asyncio.ensure_future(
                self.s.query_info(to, node="foobar")
            )

            with self.assertRaises(errors.XMPPCancelError):
                run_coroutine(task1)

            with self.assertRaises(errors.XMPPCancelError):
                run_coroutine(task2)

    def test_query_info_reraises_but_does_not_cache_exception(self):
        to = structs.JID.fromstr("user@foo.example/res1")

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            send_and_decode.side_effect = errors.XMPPCancelError(
                condition=errors.ErrorCondition.FEATURE_NOT_IMPLEMENTED
            )

            with self.assertRaises(errors.XMPPCancelError):
                run_coroutine(
                    self.s.query_info(to, node="foobar")
                )

            send_and_decode.side_effect = ConnectionError()

            with self.assertRaises(ConnectionError):
                run_coroutine(
                    self.s.query_info(to, node="foobar")
                )

    def test_query_info_cache_override(self):
        to = structs.JID.fromstr("user@foo.example/res1")

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            response1 = {}

            send_and_decode.return_value = response1

            result1 = run_coroutine(
                self.s.query_info(to, node="foobar")
            )

            response2 = {}
            send_and_decode.return_value = response2

            result2 = run_coroutine(
                self.s.query_info(to, node="foobar", require_fresh=True)
            )

        self.assertIs(result1, response1)
        self.assertIs(result2, response2)

        self.assertEqual(
            2,
            len(send_and_decode.mock_calls)
        )

    def test_flush_cache_clears_info_cache(self):
        to = structs.JID.fromstr("user@foo.example/res1")

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            response1 = {}

            send_and_decode.return_value = response1

            result1 = run_coroutine(
                self.s.query_info(to, node="foobar")
            )

            self.s.flush_cache()

            response2 = {}
            send_and_decode.return_value = response2

            result2 = run_coroutine(
                self.s.query_info(to, node="foobar")
            )

        self.assertIs(result1, response1)
        self.assertIs(result2, response2)

        self.assertEqual(
            2,
            len(send_and_decode.mock_calls)
        )

    def test_query_info_cache_clears_on_disconnect(self):
        to = structs.JID.fromstr("user@foo.example/res1")

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            response1 = {}

            send_and_decode.return_value = response1

            result1 = run_coroutine(
                self.s.query_info(to, node="foobar")
            )

            self.cc.on_stream_destroyed()

            response2 = {}
            send_and_decode.return_value = response2

            result2 = run_coroutine(
                self.s.query_info(to, node="foobar")
            )

        self.assertIs(result1, response1)
        self.assertIs(result2, response2)

        self.assertEqual(
            2,
            len(send_and_decode.mock_calls)
        )

    def test_query_info_timeout(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            response = {}

            send_and_decode.delay = 1
            send_and_decode.return_value = response

            with self.assertRaises(TimeoutError):
                result = run_coroutine(
                    self.s.query_info(to, timeout=0.01)
                )

                self.assertSequenceEqual(
                    [
                        unittest.mock.call(to, None),
                    ],
                    send_and_decode.mock_calls
                )

    def test_query_info_deduplicate_requests(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = disco_xso.InfoQuery()

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            response = {}

            send_and_decode.return_value = response

            result = run_coroutine(
                asyncio.gather(
                    self.s.query_info(to, timeout=10),
                    self.s.query_info(to, timeout=10),
                )
            )

            self.assertIs(result[0], response)
            self.assertIs(result[1], response)

        self.assertSequenceEqual(
            [
                unittest.mock.call(to, None),
            ],
            send_and_decode.mock_calls
        )

    def test_query_info_transparent_deduplication_when_cancelled(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = disco_xso.InfoQuery()

        with unittest.mock.patch.object(
                self.s,
                "send_and_decode_info_query",
                new=CoroutineMock()) as send_and_decode:
            response = {}

            send_and_decode.return_value = response
            send_and_decode.delay = 0.1

            q1 = asyncio.ensure_future(self.s.query_info(to))
            q2 = asyncio.ensure_future(self.s.query_info(to))

            run_coroutine(asyncio.sleep(0.05))

            q1.cancel()

            result = run_coroutine(q2)

        self.assertIs(result, response)

        self.assertSequenceEqual(
            [
                unittest.mock.call(to, None),
                unittest.mock.call(to, None),
            ],
            send_and_decode.mock_calls
        )

    def test_query_items(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = disco_xso.ItemsQuery()

        self.cc.send.return_value = response

        result = run_coroutine(
            self.s.query_items(to)
        )

        self.assertIs(result, response)
        self.assertEqual(
            1,
            len(self.cc.send.mock_calls)
        )

        call, = self.cc.send.mock_calls
        # call[1] are args
        request_iq, = call[1]

        self.assertEqual(
            to,
            request_iq.to
        )
        self.assertEqual(
            structs.IQType.GET,
            request_iq.type_
        )
        self.assertIsInstance(request_iq.payload, disco_xso.ItemsQuery)
        self.assertFalse(request_iq.payload.items)
        self.assertIsNone(request_iq.payload.node)

    def test_query_items_with_node(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = disco_xso.ItemsQuery()

        self.cc.send.return_value = response

        with self.assertRaises(TypeError):
            self.s.query_items(to, "foobar")

        result = run_coroutine(
            self.s.query_items(to, node="foobar")
        )

        self.assertIs(result, response)
        self.assertEqual(
            1,
            len(self.cc.send.mock_calls)
        )

        call, = self.cc.send.mock_calls
        # call[1] are args
        request_iq, = call[1]

        self.assertEqual(
            to,
            request_iq.to
        )
        self.assertEqual(
            structs.IQType.GET,
            request_iq.type_
        )
        self.assertIsInstance(request_iq.payload, disco_xso.ItemsQuery)
        self.assertFalse(request_iq.payload.items)
        self.assertEqual("foobar", request_iq.payload.node)

    def test_query_items_caches(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = disco_xso.ItemsQuery()

        self.cc.send.return_value = response

        with self.assertRaises(TypeError):
            self.s.query_items(to, "foobar")

        result1 = run_coroutine(
            self.s.query_items(to, node="foobar")
        )
        result2 = run_coroutine(
            self.s.query_items(to, node="foobar")
        )

        self.assertIs(result1, response)
        self.assertIs(result2, response)

        self.assertEqual(
            1,
            len(self.cc.send.mock_calls)
        )

    def test_query_items_reraises_and_aliases_exception(self):
        to = structs.JID.fromstr("user@foo.example/res1")

        ncall = 0

        @asyncio.coroutine
        def mock(*args, **kwargs):
            nonlocal ncall
            ncall += 1
            if ncall == 1:
                raise errors.XMPPCancelError(
                    condition=errors.ErrorCondition.FEATURE_NOT_IMPLEMENTED
                )
            else:
                raise ConnectionError()

        with unittest.mock.patch.object(
                self.cc,
                "send",
                new=mock):

            task1 = asyncio.ensure_future(
                self.s.query_info(to, node="foobar")
            )
            task2 = asyncio.ensure_future(
                self.s.query_info(to, node="foobar")
            )

            with self.assertRaises(errors.XMPPCancelError):
                run_coroutine(task1)

            with self.assertRaises(errors.XMPPCancelError):
                run_coroutine(task2)

    def test_query_info_reraises_but_does_not_cache_exception(self):
        to = structs.JID.fromstr("user@foo.example/res1")

        self.cc.send.side_effect = \
            errors.XMPPCancelError(
                condition=errors.ErrorCondition.FEATURE_NOT_IMPLEMENTED,
            )

        with self.assertRaises(errors.XMPPCancelError):
            run_coroutine(
                self.s.query_items(to, node="foobar")
            )

        self.cc.send.side_effect = \
            ConnectionError()

        with self.assertRaises(ConnectionError):
            run_coroutine(
                self.s.query_items(to, node="foobar")
            )

    def test_query_items_cache_override(self):
        to = structs.JID.fromstr("user@foo.example/res1")

        response1 = disco_xso.ItemsQuery()
        self.cc.send.return_value = response1

        with self.assertRaises(TypeError):
            self.s.query_items(to, "foobar")

        result1 = run_coroutine(
            self.s.query_items(to, node="foobar")
        )

        response2 = disco_xso.ItemsQuery()
        self.cc.send.return_value = response2

        result2 = run_coroutine(
            self.s.query_items(to, node="foobar", require_fresh=True)
        )

        self.assertIs(result1, response1)
        self.assertIs(result2, response2)

        self.assertEqual(
            2,
            len(self.cc.send.mock_calls)
        )

    def test_flush_cache_clears_items_cache(self):
        to = structs.JID.fromstr("user@foo.example/res1")

        response1 = disco_xso.ItemsQuery()
        self.cc.send.return_value = response1

        with self.assertRaises(TypeError):
            self.s.query_items(to, "foobar")

        result1 = run_coroutine(
            self.s.query_items(to, node="foobar")
        )

        response2 = disco_xso.ItemsQuery()
        self.cc.send.return_value = response2

        self.s.flush_cache()

        result2 = run_coroutine(
            self.s.query_items(to, node="foobar")
        )

        self.assertIs(result1, response1)
        self.assertIs(result2, response2)

        self.assertEqual(
            2,
            len(self.cc.send.mock_calls)
        )

    def test_query_items_cache_clears_on_disconnect(self):
        to = structs.JID.fromstr("user@foo.example/res1")

        response1 = disco_xso.ItemsQuery()
        self.cc.send.return_value = response1

        with self.assertRaises(TypeError):
            self.s.query_items(to, "foobar")

        result1 = run_coroutine(
            self.s.query_items(to, node="foobar")
        )

        self.cc.on_stream_destroyed()

        response2 = disco_xso.ItemsQuery()
        self.cc.send.return_value = response2

        result2 = run_coroutine(
            self.s.query_items(to, node="foobar")
        )

        self.assertIs(result1, response1)
        self.assertIs(result2, response2)

        self.assertEqual(
            2,
            len(self.cc.send.mock_calls)
        )

    def test_query_items_timeout(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = disco_xso.ItemsQuery()

        self.cc.send.delay = 1
        self.cc.send.return_value = response

        with self.assertRaises(TimeoutError):
            result = run_coroutine(
                self.s.query_items(to, timeout=0.01)
            )

        self.assertSequenceEqual(
            [
                unittest.mock.call(unittest.mock.ANY),
            ],
            self.cc.send.mock_calls
        )

    def test_query_items_deduplicate_requests(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = disco_xso.ItemsQuery()

        self.cc.send.return_value = response

        result = run_coroutine(
            asyncio.gather(
                self.s.query_items(to, timeout=10),
                self.s.query_items(to, timeout=10),
            )
        )

        self.assertIs(result[0], response)
        self.assertIs(result[1], response)

        self.assertSequenceEqual(
            [
                unittest.mock.call(unittest.mock.ANY),
            ],
            self.cc.send.mock_calls
        )

    def test_query_items_transparent_deduplication_when_cancelled(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = disco_xso.ItemsQuery()

        self.cc.send.return_value = response
        self.cc.send.delay = 0.1

        q1 = asyncio.ensure_future(self.s.query_items(to))
        q2 = asyncio.ensure_future(self.s.query_items(to))

        run_coroutine(asyncio.sleep(0.05))

        q1.cancel()

        result = run_coroutine(q2)

        self.assertIs(result, response)

        self.assertSequenceEqual(
            [
                unittest.mock.call(unittest.mock.ANY),
                unittest.mock.call(unittest.mock.ANY),
            ],
            self.cc.send.mock_calls
        )

    def test_set_info_cache(self):
        to = structs.JID.fromstr("user@foo.example/res1")
        response = disco_xso.ItemsQuery()

        self.s.set_info_cache(
            to,
            None,
            response
        )

        other_response = disco_xso.InfoQuery()
        self.cc.send.return_value = \
            other_response

        result = run_coroutine(self.s.query_info(to, node=None))

        self.assertIs(result, response)
        self.assertFalse(self.cc.stream.mock_calls)

    def test_set_info_future(self):
        to = structs.JID.fromstr("user@foo.example/res1")

        fut = asyncio.Future()

        self.s.set_info_future(
            to,
            None,
            fut
        )

        request = asyncio.ensure_future(
            self.s.query_info(to)
        )

        run_coroutine(asyncio.sleep(0))
        self.assertFalse(request.done())

        result = object()
        fut.set_result(result)

        self.assertIs(run_coroutine(request), result)

    def test_set_info_future_stays_even_with_exception(self):
        exc = ConnectionError()
        to = structs.JID.fromstr("user@foo.example/res1")

        fut = asyncio.Future()

        self.s.set_info_future(
            to,
            None,
            fut
        )

        request = asyncio.ensure_future(
            self.s.query_info(to)
        )

        run_coroutine(asyncio.sleep(0))
        self.assertFalse(request.done())

        fut.set_exception(exc)

        with self.assertRaises(Exception) as ctx:
            run_coroutine(request)

        self.assertIs(ctx.exception, exc)


class Testmount_as_node(unittest.TestCase):
    def setUp(self):
        self.pn = disco_service.mount_as_node(
            unittest.mock.sentinel.mountpoint
        )
        self.instance = unittest.mock.Mock()
        self.disco_server = unittest.mock.Mock()
        self.instance.dependencies = {
            disco_service.DiscoServer: self.disco_server
        }

    def tearDown(self):
        del self.instance
        del self.disco_server
        del self.pn

    def test_value_type(self):
        self.assertIs(
            self.pn.value_type,
            type(None),
        )

    def test_mountpoint(self):
        self.assertEqual(
            self.pn.mountpoint,
            unittest.mock.sentinel.mountpoint,
        )

    def test_required_dependencies(self):
        self.assertSetEqual(
            set(self.pn.required_dependencies),
            {disco_service.DiscoServer},
        )

    def test_contextmanager(self):
        cm = self.pn.init_cm(self.instance)
        self.disco_server.mount_node.assert_not_called()
        cm.__enter__()
        self.disco_server.mount_node.assert_called_once_with(
            unittest.mock.sentinel.mountpoint,
            self.instance,
        )
        self.disco_server.unmount_node.assert_not_called()
        cm.__exit__(None, None, None)
        self.disco_server.unmount_node.assert_called_once_with(
            unittest.mock.sentinel.mountpoint,
        )

    def test_contextmanager_is_exception_safe(self):
        cm = self.pn.init_cm(self.instance)
        self.disco_server.mount_node.assert_not_called()
        cm.__enter__()
        self.disco_server.mount_node.assert_called_once_with(
            unittest.mock.sentinel.mountpoint,
            self.instance,
        )
        self.disco_server.unmount_node.assert_not_called()
        try:
            raise Exception()
        except:  # NOQA
            cm.__exit__(*sys.exc_info())
        self.disco_server.unmount_node.assert_called_once_with(
            unittest.mock.sentinel.mountpoint,
        )


class TestRegisteredFeature(unittest.TestCase):
    def setUp(self):
        self.s = unittest.mock.Mock()
        self.rf = disco_service.RegisteredFeature(
            self.s,
            unittest.mock.sentinel.feature,
        )

    def test_contextmanager(self):
        self.s.register_feature.assert_not_called()
        result = self.rf.__enter__()
        self.assertIs(result, self.rf)
        self.s.register_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )
        self.s.unregister_feature.assert_not_called()
        result = self.rf.__exit__(None, None, None)
        self.assertFalse(result)
        self.s.unregister_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )

    def test_contextmanager_is_exception_safe(self):
        class FooException(Exception):
            pass

        self.s.register_feature.assert_not_called()
        with self.assertRaises(FooException):
            with self.rf:
                self.s.register_feature.assert_called_once_with(
                    unittest.mock.sentinel.feature,
                )
                self.s.unregister_feature.assert_not_called()

                raise FooException()

        self.s.unregister_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )

    def test_feature(self):
        self.assertEqual(
            self.rf.feature,
            unittest.mock.sentinel.feature,
        )

    def test_feature_is_not_writable(self):
        with self.assertRaises(AttributeError):
            self.rf.feature = self.rf.feature

    def test_enabled(self):
        self.assertFalse(self.rf.enabled)

    def test_enabled_changes_with_cm_use(self):
        with self.rf:
            self.assertTrue(self.rf.enabled)
        self.assertFalse(self.rf.enabled)

    def test_setting_enabled_registers_feature(self):
        self.assertFalse(self.rf.enabled)
        self.s.register_feature.assert_not_called()
        self.rf.enabled = True
        self.assertTrue(self.rf.enabled)
        self.s.register_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )

    def test_setting_enabled_is_idempotent(self):
        self.assertFalse(self.rf.enabled)
        self.s.register_feature.assert_not_called()
        self.rf.enabled = True
        self.assertTrue(self.rf.enabled)
        self.s.register_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )

        self.rf.enabled = True
        self.assertTrue(self.rf.enabled)
        self.s.register_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )

    def test_clearing_enabled_unregisters_feature(self):
        self.assertFalse(self.rf.enabled)
        self.s.register_feature.assert_not_called()
        self.rf.enabled = True
        self.s.register_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )
        self.s.unregister_feature.assert_not_called()

        self.rf.enabled = False

        self.s.unregister_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )

    def test_clearing_enabled_is_idempotent(self):
        self.assertFalse(self.rf.enabled)
        self.s.register_feature.assert_not_called()
        self.rf.enabled = True
        self.s.register_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )
        self.s.unregister_feature.assert_not_called()

        self.rf.enabled = False
        self.assertFalse(self.rf.enabled)

        self.s.unregister_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )

        self.rf.enabled = False

        self.s.unregister_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )

    def test_clearing_enabled_while_in_cm_does_not_duplicate_unregister(self):
        with self.rf:
            self.rf.enabled = False

            self.s.unregister_feature.assert_called_once_with(
                unittest.mock.sentinel.feature,
            )

        self.s.unregister_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )

    def test_setting_enabled_before_entering_cm_does_not_duplicate_register(self):  # NOQA
        self.rf.enabled = True

        self.s.register_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )

        with self.rf:
            self.s.register_feature.assert_called_once_with(
                unittest.mock.sentinel.feature,
            )


class Testregister_feature(unittest.TestCase):
    def setUp(self):
        self.pn = disco_service.register_feature(
            unittest.mock.sentinel.feature
        )
        self.instance = unittest.mock.Mock()
        self.disco_server = unittest.mock.Mock()
        self.instance.dependencies = {
            disco_service.DiscoServer: self.disco_server
        }

    def tearDown(self):
        del self.instance
        del self.disco_server
        del self.pn

    def test_value_type(self):
        self.assertIs(
            self.pn.value_type,
            disco_service.RegisteredFeature,
        )

    def test_mountpoint(self):
        self.assertEqual(
            self.pn.feature,
            unittest.mock.sentinel.feature,
        )

    def test_required_dependencies(self):
        self.assertSetEqual(
            set(self.pn.required_dependencies),
            {disco_service.DiscoServer},
        )

    def test_init_cm_creates_RegisteredFeature(self):
        with contextlib.ExitStack() as stack:
            RegisteredFeature = stack.enter_context(
                unittest.mock.patch("aioxmpp.disco.service.RegisteredFeature")
            )

            result = self.pn.init_cm(self.instance)

        RegisteredFeature.assert_called_once_with(
            self.disco_server,
            unittest.mock.sentinel.feature,
        )

        self.assertEqual(result, RegisteredFeature())

    def test_contextmanager(self):
        cm = self.pn.init_cm(self.instance)
        self.disco_server.register_feature.assert_not_called()
        cm.__enter__()
        self.disco_server.register_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )
        self.disco_server.unregister_feature.assert_not_called()
        cm.__exit__(None, None, None)
        self.disco_server.unregister_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )

    def test_contextmanager_is_exception_safe(self):
        cm = self.pn.init_cm(self.instance)
        self.disco_server.register_feature.assert_not_called()
        cm.__enter__()
        self.disco_server.register_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )
        self.disco_server.unregister_feature.assert_not_called()
        try:
            raise Exception()
        except:  # NOQA
            cm.__exit__(*sys.exc_info())
        self.disco_server.unregister_feature.assert_called_once_with(
            unittest.mock.sentinel.feature,
        )



########################################################################
# File name: test_connector.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import contextlib
import logging
import unittest
import unittest.mock

import aioxmpp.connector as connector
import aioxmpp.errors as errors
import aioxmpp.nonza as nonza

from aioxmpp.utils import namespaces

from aioxmpp.testutils import (
    run_coroutine,
    CoroutineMock,
)


class TestSTARTTLSConnector(unittest.TestCase):
    def setUp(self):
        self.c = connector.STARTTLSConnector()

    def tearDown(self):
        del self.c

    def test_tls_supported(self):
        self.assertTrue(
            self.c.tls_supported
        )

    def test_connect_successful(self):
        captured_features_future = None

        def capture_future(*args, features_future=None, **kwargs):
            nonlocal captured_features_future
            captured_features_future = features_future
            return base.protocol

        features = nonza.StreamFeatures()
        features[...] = nonza.StartTLSFeature()

        features_future = asyncio.Future()
        features_future.set_result(
            features
        )

        base_logger = unittest.mock.Mock(spec=logging.Logger)

        base = unittest.mock.Mock()
        base.protocol.starttls = CoroutineMock()
        base.create_starttls_connection = CoroutineMock()
        base.create_starttls_connection.return_value = (
            unittest.mock.sentinel.transport,
            base.protocol,
        )
        base.metadata.tls_required = True
        base.XMLStream.return_value = base.protocol
        base.XMLStream.side_effect = capture_future
        base.Future.return_value = features_future
        base.send_and_wait_for = CoroutineMock()
        base.send_and_wait_for.return_value = unittest.mock.Mock(
            spec=nonza.StartTLSProceed,
        )
        base.certificate_verifier.pre_handshake = CoroutineMock()
        base.metadata.certificate_verifier_factory.return_value = \
            base.certificate_verifier
        base.metadata.ssl_context_factory.return_value = \
            unittest.mock.sentinel.ssl_context
        base.reset_stream_and_get_features = CoroutineMock()
        base.reset_stream_and_get_features.return_value = \
            unittest.mock.sentinel.reset
        base.async_.return_value = unittest.mock.sentinel.features_future

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.Future",
                    new=base.Future,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.ssl_transport.create_starttls_connection",
                    new=base.create_starttls_connection,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.XMLStream",
                    new=base.XMLStream,
                )
            )

            StartTLS_nonza = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.nonza.StartTLS",
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.send_and_wait_for",
                    new=base.send_and_wait_for,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.reset_stream_and_get_features",
                    new=base.reset_stream_and_get_features,
                )
            )

            timedelta = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.connector.timedelta",
                )
            )

            result = run_coroutine(self.c.connect(
                unittest.mock.sentinel.loop,
                base.metadata,
                unittest.mock.sentinel.domain,
                unittest.mock.sentinel.host,
                unittest.mock.sentinel.port,
                unittest.mock.sentinel.timeout,
                base_logger=base_logger,
            ))

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.Future(
                    loop=unittest.mock.sentinel.loop,
                ),
                unittest.mock.call.XMLStream(
                    to=unittest.mock.sentinel.domain,
                    features_future=features_future,
                    base_logger=base_logger,
                ),
                unittest.mock.call.create_starttls_connection(
                    unittest.mock.sentinel.loop,
                    unittest.mock.ANY,
                    host=unittest.mock.sentinel.host,
                    port=unittest.mock.sentinel.port,
                    peer_hostname=unittest.mock.sentinel.host,
                    server_hostname=unittest.mock.sentinel.domain,
                    use_starttls=True,
                ),
                unittest.mock.call.send_and_wait_for(
                    base.protocol,
                    [
                        StartTLS_nonza(),
                    ],
                    [
                        nonza.StartTLSFailure,
                        nonza.StartTLSProceed,
                    ]
                ),
                unittest.mock.call.metadata.certificate_verifier_factory(),
                unittest.mock.call.certificate_verifier.pre_handshake(
                    unittest.mock.sentinel.domain,
                    unittest.mock.sentinel.host,
                    unittest.mock.sentinel.port,
                    base.metadata,
                ),
                unittest.mock.call.metadata.ssl_context_factory(),
                unittest.mock.call.certificate_verifier.setup_context(
                    unittest.mock.sentinel.ssl_context,
                    unittest.mock.sentinel.transport,
                ),
                unittest.mock.call.protocol.starttls(
                    ssl_context=unittest.mock.sentinel.ssl_context,
                    post_handshake_callback=
                        base.certificate_verifier.post_handshake,
                ),
                unittest.mock.call.reset_stream_and_get_features(
                    base.protocol,
                    timeout=unittest.mock.sentinel.timeout,
                ),
            ]
        )

        self.assertEqual(
            result,
            (
                unittest.mock.sentinel.transport,
                base.protocol,
                unittest.mock.sentinel.reset,
            )
        )

        timedelta.assert_called_once_with(
            seconds=unittest.mock.sentinel.timeout
        )

        self.assertEqual(
            base.protocol.deadtime_hard_limit,
            timedelta(),
        )

    def test_abort_xmlstream_if_connect_fails(self):
        captured_features_future = None

        def capture_future(*args, features_future=None, **kwargs):
            nonlocal captured_features_future
            captured_features_future = features_future
            return base.protocol

        features = nonza.StreamFeatures()
        features[...] = nonza.StartTLSFeature()

        features_future = asyncio.Future()
        features_future.set_result(
            features
        )

        base = unittest.mock.Mock()
        base.create_starttls_connection = CoroutineMock()
        base.create_starttls_connection.side_effect = Exception()
        base.XMLStream.return_value = base.protocol
        base.XMLStream.side_effect = capture_future
        base.Future.return_value = features_future

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.Future",
                    new=base.Future,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.ssl_transport.create_starttls_connection",
                    new=base.create_starttls_connection,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.XMLStream",
                    new=base.XMLStream,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.send_and_wait_for",
                    new=base.send_and_wait_for,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.reset_stream_and_get_features",
                    new=base.reset_stream_and_get_features,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.connector.timedelta",
                )
            )

            with self.assertRaises(Exception):
                run_coroutine(self.c.connect(
                    unittest.mock.sentinel.loop,
                    base.metadata,
                    unittest.mock.sentinel.domain,
                    unittest.mock.sentinel.host,
                    unittest.mock.sentinel.port,
                    unittest.mock.sentinel.timeout,
                ))

        self.maxDiff = None

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.Future(
                    loop=unittest.mock.sentinel.loop,
                ),
                unittest.mock.call.XMLStream(
                    to=unittest.mock.sentinel.domain,
                    features_future=features_future,
                    base_logger=None,
                ),
                unittest.mock.call.create_starttls_connection(
                    unittest.mock.sentinel.loop,
                    unittest.mock.ANY,
                    host=unittest.mock.sentinel.host,
                    port=unittest.mock.sentinel.port,
                    peer_hostname=unittest.mock.sentinel.host,
                    server_hostname=unittest.mock.sentinel.domain,
                    use_starttls=True,
                ),
                unittest.mock.call.protocol.abort(),
            ]
        )

    def test_connect_without_starttls_support_and_with_required_success(self):
        captured_features_future = None

        def capture_future(*args, features_future=None, **kwargs):
            nonlocal captured_features_future
            captured_features_future = features_future
            return base.protocol

        features = nonza.StreamFeatures()

        features_future = asyncio.Future()
        features_future.set_result(
            features
        )

        base = unittest.mock.Mock()
        base.protocol.starttls = CoroutineMock()
        base.create_starttls_connection = CoroutineMock()
        base.create_starttls_connection.return_value = (
            unittest.mock.sentinel.transport,
            base.protocol,
        )
        base.metadata.tls_required = True
        base.XMLStream.return_value = base.protocol
        base.XMLStream.side_effect = capture_future
        base.Future.return_value = features_future
        base.send_and_wait_for = CoroutineMock()
        base.send_and_wait_for.return_value = unittest.mock.Mock(
            spec=nonza.StartTLSProceed,
        )
        base.certificate_verifier.pre_handshake = CoroutineMock()
        base.metadata.certificate_verifier_factory.return_value = \
            base.certificate_verifier
        base.metadata.ssl_context_factory.return_value = \
            unittest.mock.sentinel.ssl_context
        base.reset_stream_and_get_features = CoroutineMock()
        base.reset_stream_and_get_features.return_value = \
            unittest.mock.sentinel.reset
        base.async_.return_value = unittest.mock.sentinel.features_future

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.Future",
                    new=base.Future,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.ssl_transport.create_starttls_connection",
                    new=base.create_starttls_connection,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.XMLStream",
                    new=base.XMLStream,
                )
            )

            StartTLS_nonza = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.nonza.StartTLS",
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.send_and_wait_for",
                    new=base.send_and_wait_for,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.reset_stream_and_get_features",
                    new=base.reset_stream_and_get_features,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.connector.timedelta",
                )
            )

            result = run_coroutine(self.c.connect(
                unittest.mock.sentinel.loop,
                base.metadata,
                unittest.mock.sentinel.domain,
                unittest.mock.sentinel.host,
                unittest.mock.sentinel.port,
                unittest.mock.sentinel.timeout,
            ))

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.Future(
                    loop=unittest.mock.sentinel.loop,
                ),
                unittest.mock.call.XMLStream(
                    to=unittest.mock.sentinel.domain,
                    features_future=features_future,
                    base_logger=None,
                ),
                unittest.mock.call.create_starttls_connection(
                    unittest.mock.sentinel.loop,
                    unittest.mock.ANY,
                    host=unittest.mock.sentinel.host,
                    port=unittest.mock.sentinel.port,
                    peer_hostname=unittest.mock.sentinel.host,
                    server_hostname=unittest.mock.sentinel.domain,
                    use_starttls=True,
                ),
                unittest.mock.call.send_and_wait_for(
                    base.protocol,
                    [
                        StartTLS_nonza(),
                    ],
                    [
                        nonza.StartTLSFailure,
                        nonza.StartTLSProceed,
                    ]
                ),
                unittest.mock.call.metadata.certificate_verifier_factory(),
                unittest.mock.call.certificate_verifier.pre_handshake(
                    unittest.mock.sentinel.domain,
                    unittest.mock.sentinel.host,
                    unittest.mock.sentinel.port,
                    base.metadata,
                ),
                unittest.mock.call.metadata.ssl_context_factory(),
                unittest.mock.call.certificate_verifier.setup_context(
                    unittest.mock.sentinel.ssl_context,
                    unittest.mock.sentinel.transport,
                ),
                unittest.mock.call.protocol.starttls(
                    ssl_context=unittest.mock.sentinel.ssl_context,
                    post_handshake_callback=
                        base.certificate_verifier.post_handshake,
                ),
                unittest.mock.call.reset_stream_and_get_features(
                    base.protocol,
                    timeout=unittest.mock.sentinel.timeout,
                ),
            ]
        )

        self.assertEqual(
            result,
            (
                unittest.mock.sentinel.transport,
                base.protocol,
                unittest.mock.sentinel.reset,
            )
        )

    def test_connect_without_starttls_support_and_with_required_failure(self):
        captured_features_future = None

        def capture_future(*args, features_future=None, **kwargs):
            nonlocal captured_features_future
            captured_features_future = features_future
            return base.protocol

        features = nonza.StreamFeatures()

        features_future = asyncio.Future()
        features_future.set_result(
            features
        )

        base = unittest.mock.Mock()
        base.protocol.starttls = CoroutineMock()
        base.create_starttls_connection = CoroutineMock()
        base.create_starttls_connection.return_value = (
            unittest.mock.sentinel.transport,
            base.protocol,
        )
        base.metadata.tls_required = True
        base.XMLStream.return_value = base.protocol
        base.XMLStream.side_effect = capture_future
        base.Future.return_value = features_future
        base.send_and_wait_for = CoroutineMock()
        base.send_and_wait_for.return_value = unittest.mock.Mock(
            spec=nonza.StartTLSFailure,
        )

        error_message = (
            "server failed to STARTTLS"
        )

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.Future",
                    new=base.Future,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.ssl_transport.create_starttls_connection",
                    new=base.create_starttls_connection,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.XMLStream",
                    new=base.XMLStream,
                )
            )

            StartTLS_nonza = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.nonza.StartTLS",
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.send_and_wait_for",
                    new=base.send_and_wait_for,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.send_stream_error_and_close",
                    new=base.send_stream_error_and_close
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.ensure_future",
                    new=base.async_,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.connector.timedelta",
                )
            )

            with self.assertRaisesRegex(errors.TLSUnavailable, error_message):
                run_coroutine(self.c.connect(
                    unittest.mock.sentinel.loop,
                    base.metadata,
                    unittest.mock.sentinel.domain,
                    unittest.mock.sentinel.host,
                    unittest.mock.sentinel.port,
                    unittest.mock.sentinel.timeout,
                ))

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.Future(
                    loop=unittest.mock.sentinel.loop,
                ),
                unittest.mock.call.XMLStream(
                    to=unittest.mock.sentinel.domain,
                    features_future=features_future,
                    base_logger=None,
                ),
                unittest.mock.call.create_starttls_connection(
                    unittest.mock.sentinel.loop,
                    unittest.mock.ANY,
                    host=unittest.mock.sentinel.host,
                    port=unittest.mock.sentinel.port,
                    peer_hostname=unittest.mock.sentinel.host,
                    server_hostname=unittest.mock.sentinel.domain,
                    use_starttls=True,
                ),
                unittest.mock.call.send_and_wait_for(
                    base.protocol,
                    [
                        StartTLS_nonza(),
                    ],
                    [
                        nonza.StartTLSFailure,
                        nonza.StartTLSProceed,
                    ]
                ),
                unittest.mock.call.send_stream_error_and_close(
                    base.protocol,
                    condition=errors.StreamErrorCondition.POLICY_VIOLATION,
                    text=error_message,
                )
            ]
        )

    def test_connect_without_starttls_support_and_with_required_error(self):
        captured_features_future = None

        def capture_future(*args, features_future=None, **kwargs):
            nonlocal captured_features_future
            captured_features_future = features_future
            return base.protocol

        features = nonza.StreamFeatures()

        features_future = asyncio.Future()
        features_future.set_result(
            features
        )

        base = unittest.mock.Mock()
        base.protocol.starttls = CoroutineMock()
        base.create_starttls_connection = CoroutineMock()
        base.create_starttls_connection.return_value = (
            unittest.mock.sentinel.transport,
            base.protocol,
        )
        base.metadata.tls_required = True
        base.XMLStream.return_value = base.protocol
        base.XMLStream.side_effect = capture_future
        base.Future.return_value = features_future
        base.send_and_wait_for = CoroutineMock()
        base.send_and_wait_for.side_effect = errors.StreamError(
            condition=errors.StreamErrorCondition.UNSUPPORTED_STANZA_TYPE,
        )

        error_message = (
            "STARTTLS not supported by server, but required by client"
        )

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.Future",
                    new=base.Future,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.ssl_transport.create_starttls_connection",
                    new=base.create_starttls_connection,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.XMLStream",
                    new=base.XMLStream,
                )
            )

            StartTLS_nonza = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.nonza.StartTLS",
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.send_and_wait_for",
                    new=base.send_and_wait_for,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.send_stream_error_and_close",
                    new=base.send_stream_error_and_close
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.ensure_future",
                    new=base.async_,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.connector.timedelta",
                )
            )

            with self.assertRaisesRegex(errors.TLSUnavailable, error_message):
                run_coroutine(self.c.connect(
                    unittest.mock.sentinel.loop,
                    base.metadata,
                    unittest.mock.sentinel.domain,
                    unittest.mock.sentinel.host,
                    unittest.mock.sentinel.port,
                    unittest.mock.sentinel.timeout,
                ))

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.Future(
                    loop=unittest.mock.sentinel.loop,
                ),
                unittest.mock.call.XMLStream(
                    to=unittest.mock.sentinel.domain,
                    features_future=features_future,
                    base_logger=None,
                ),
                unittest.mock.call.create_starttls_connection(
                    unittest.mock.sentinel.loop,
                    unittest.mock.ANY,
                    host=unittest.mock.sentinel.host,
                    port=unittest.mock.sentinel.port,
                    peer_hostname=unittest.mock.sentinel.host,
                    server_hostname=unittest.mock.sentinel.domain,
                    use_starttls=True,
                ),
                unittest.mock.call.send_and_wait_for(
                    base.protocol,
                    [
                        StartTLS_nonza(),
                    ],
                    [
                        nonza.StartTLSFailure,
                        nonza.StartTLSProceed,
                    ]
                )
            ]
        )

    def test_connect_without_starttls_support_and_without_required(self):
        captured_features_future = None

        def capture_future(*args, features_future=None, **kwargs):
            nonlocal captured_features_future
            captured_features_future = features_future
            return base.protocol

        features = nonza.StreamFeatures()

        features_future = asyncio.Future()
        features_future.set_result(
            features
        )

        base = unittest.mock.Mock()
        base.protocol.starttls = CoroutineMock()
        base.create_starttls_connection = CoroutineMock()
        base.create_starttls_connection.return_value = (
            unittest.mock.sentinel.transport,
            base.protocol,
        )
        base.metadata.tls_required = False
        base.XMLStream.return_value = base.protocol
        base.XMLStream.side_effect = capture_future
        base.Future.return_value = features_future

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.Future",
                    new=base.Future,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.ssl_transport.create_starttls_connection",
                    new=base.create_starttls_connection,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.XMLStream",
                    new=base.XMLStream,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.send_stream_error_and_close",
                    new=base.send_stream_error_and_close,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.connector.timedelta",
                )
            )

            result = run_coroutine(self.c.connect(
                unittest.mock.sentinel.loop,
                base.metadata,
                unittest.mock.sentinel.domain,
                unittest.mock.sentinel.host,
                unittest.mock.sentinel.port,
                unittest.mock.sentinel.timeout,
            ))

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.Future(
                    loop=unittest.mock.sentinel.loop,
                ),
                unittest.mock.call.XMLStream(
                    to=unittest.mock.sentinel.domain,
                    features_future=features_future,
                    base_logger=None,
                ),
                unittest.mock.call.create_starttls_connection(
                    unittest.mock.sentinel.loop,
                    unittest.mock.ANY,
                    host=unittest.mock.sentinel.host,
                    port=unittest.mock.sentinel.port,
                    peer_hostname=unittest.mock.sentinel.host,
                    server_hostname=unittest.mock.sentinel.domain,
                    use_starttls=True,
                ),
            ]
        )

        self.assertEqual(
            result,
            (
                unittest.mock.sentinel.transport,
                base.protocol,
                features,
            )
        )

    def test_connect_with_failed_starttls_and_with_required(self):
        captured_features_future = None

        def capture_future(*args, features_future=None, **kwargs):
            nonlocal captured_features_future
            captured_features_future = features_future
            return base.protocol

        features = nonza.StreamFeatures()
        features[...] = nonza.StartTLSFeature()

        features_future = asyncio.Future()
        features_future.set_result(
            features
        )

        base = unittest.mock.Mock()
        base.protocol.starttls = CoroutineMock()
        base.create_starttls_connection = CoroutineMock()
        base.create_starttls_connection.return_value = (
            unittest.mock.sentinel.transport,
            base.protocol,
        )
        base.metadata.tls_required = True
        base.XMLStream.return_value = base.protocol
        base.XMLStream.side_effect = capture_future
        base.Future.return_value = features_future
        base.send_and_wait_for = CoroutineMock()
        base.send_and_wait_for.return_value = unittest.mock.Mock(
            spec=nonza.StartTLSFailure,
        )

        error_message = (
            "server failed to STARTTLS"
        )

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.Future",
                    new=base.Future,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.ssl_transport.create_starttls_connection",
                    new=base.create_starttls_connection,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.XMLStream",
                    new=base.XMLStream,
                )
            )

            StartTLS_nonza = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.nonza.StartTLS",
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.send_and_wait_for",
                    new=base.send_and_wait_for,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.send_stream_error_and_close",
                    new=base.send_stream_error_and_close
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.ensure_future",
                    new=base.async_,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.connector.timedelta",
                )
            )

            with self.assertRaisesRegex(errors.TLSUnavailable, error_message):
                run_coroutine(self.c.connect(
                    unittest.mock.sentinel.loop,
                    base.metadata,
                    unittest.mock.sentinel.domain,
                    unittest.mock.sentinel.host,
                    unittest.mock.sentinel.port,
                    unittest.mock.sentinel.timeout,
                ))

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.Future(
                    loop=unittest.mock.sentinel.loop,
                ),
                unittest.mock.call.XMLStream(
                    to=unittest.mock.sentinel.domain,
                    features_future=features_future,
                    base_logger=None,
                ),
                unittest.mock.call.create_starttls_connection(
                    unittest.mock.sentinel.loop,
                    unittest.mock.ANY,
                    host=unittest.mock.sentinel.host,
                    port=unittest.mock.sentinel.port,
                    peer_hostname=unittest.mock.sentinel.host,
                    server_hostname=unittest.mock.sentinel.domain,
                    use_starttls=True,
                ),
                unittest.mock.call.send_and_wait_for(
                    base.protocol,
                    [
                        StartTLS_nonza(),
                    ],
                    [
                        nonza.StartTLSFailure,
                        nonza.StartTLSProceed,
                    ]
                ),
                unittest.mock.call.send_stream_error_and_close(
                    base.protocol,
                    condition=errors.StreamErrorCondition.POLICY_VIOLATION,
                    text=error_message,
                )
            ]
        )

    def test_connect_with_failed_starttls_and_without_required(self):
        captured_features_future = None

        def capture_future(*args, features_future=None, **kwargs):
            nonlocal captured_features_future
            captured_features_future = features_future
            return base.protocol

        features = nonza.StreamFeatures()
        features[...] = nonza.StartTLSFeature()

        features_future = asyncio.Future()
        features_future.set_result(
            features
        )

        base = unittest.mock.Mock()
        base.protocol.starttls = CoroutineMock()
        base.create_starttls_connection = CoroutineMock()
        base.create_starttls_connection.return_value = (
            unittest.mock.sentinel.transport,
            base.protocol,
        )
        base.metadata.tls_required = False
        base.XMLStream.return_value = base.protocol
        base.XMLStream.side_effect = capture_future
        base.Future.return_value = features_future
        base.send_and_wait_for = CoroutineMock()
        base.send_and_wait_for.return_value = unittest.mock.Mock(
            spec=nonza.StartTLSFailure,
        )

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.Future",
                    new=base.Future,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.ssl_transport.create_starttls_connection",
                    new=base.create_starttls_connection,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.XMLStream",
                    new=base.XMLStream,
                )
            )

            StartTLS_nonza = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.nonza.StartTLS",
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.send_and_wait_for",
                    new=base.send_and_wait_for,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.send_stream_error_and_close",
                    new=base.send_stream_error_and_close
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.ensure_future",
                    new=base.async_,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.connector.timedelta",
                )
            )

            result = run_coroutine(self.c.connect(
                unittest.mock.sentinel.loop,
                base.metadata,
                unittest.mock.sentinel.domain,
                unittest.mock.sentinel.host,
                unittest.mock.sentinel.port,
                unittest.mock.sentinel.timeout,
            ))

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.Future(
                    loop=unittest.mock.sentinel.loop,
                ),
                unittest.mock.call.XMLStream(
                    to=unittest.mock.sentinel.domain,
                    features_future=features_future,
                    base_logger=None,
                ),
                unittest.mock.call.create_starttls_connection(
                    unittest.mock.sentinel.loop,
                    unittest.mock.ANY,
                    host=unittest.mock.sentinel.host,
                    port=unittest.mock.sentinel.port,
                    peer_hostname=unittest.mock.sentinel.host,
                    server_hostname=unittest.mock.sentinel.domain,
                    use_starttls=True,
                ),
                unittest.mock.call.send_and_wait_for(
                    base.protocol,
                    [
                        StartTLS_nonza(),
                    ],
                    [
                        nonza.StartTLSFailure,
                        nonza.StartTLSProceed,
                    ]
                )
            ]
        )

        self.assertEqual(
            result,
            (
                unittest.mock.sentinel.transport,
                base.protocol,
                features,
            )
        )


class TestXMPPOverTLSConnector(unittest.TestCase):
    def setUp(self):
        self.c = connector.XMPPOverTLSConnector()

    def tearDown(self):
        del self.c

    def test_tls_supported(self):
        self.assertTrue(
            self.c.tls_supported
        )

    def test_connect_with_tls(self):
        captured_features_future = None

        def capture_future(*args, features_future=None, **kwargs):
            nonlocal captured_features_future
            captured_features_future = features_future
            return base.protocol

        features_future = asyncio.Future()
        features_future.set_result(
            unittest.mock.sentinel.features
        )

        base = unittest.mock.Mock()
        base.protocol.starttls = CoroutineMock()
        base.create_starttls_connection = CoroutineMock()
        base.create_starttls_connection.return_value = (
            unittest.mock.sentinel.transport,
            base.protocol,
        )
        base.metadata.tls_required = True
        base.XMLStream.return_value = base.protocol
        base.XMLStream.side_effect = capture_future
        base.Future.return_value = features_future
        base.certificate_verifier.pre_handshake = CoroutineMock()
        base.metadata.certificate_verifier_factory.return_value = \
            base.certificate_verifier

        base._context_factory_factory.return_value = \
            unittest.mock.sentinel.ssl_context_factory

        base_logger = unittest.mock.Mock(spec=logging.Logger)

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.Future",
                    new=base.Future,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.ssl_transport.create_starttls_connection",
                    new=base.create_starttls_connection,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.XMLStream",
                    new=base.XMLStream,
                )
            )

            stack.enter_context(
                unittest.mock.patch.object(
                    self.c,
                    "_context_factory_factory",
                    new=base._context_factory_factory,
                )
            )

            timedelta = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.connector.timedelta",
                )
            )

            result = run_coroutine(self.c.connect(
                unittest.mock.sentinel.loop,
                base.metadata,
                unittest.mock.sentinel.domain,
                unittest.mock.sentinel.host,
                unittest.mock.sentinel.port,
                unittest.mock.sentinel.timeout,
                base_logger=base_logger,
            ))

        self.assertSequenceEqual(
            base_logger.mock_calls,
            [unittest.mock.call.getChild('XMPPOverTLSConnector')]
        )

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.Future(
                    loop=unittest.mock.sentinel.loop,
                ),
                unittest.mock.call.XMLStream(
                    to=unittest.mock.sentinel.domain,
                    features_future=features_future,
                    base_logger=base_logger,
                ),
                unittest.mock.call.metadata.certificate_verifier_factory(),
                unittest.mock.call.certificate_verifier.pre_handshake(
                    unittest.mock.sentinel.domain,
                    unittest.mock.sentinel.host,
                    unittest.mock.sentinel.port,
                    base.metadata,
                ),
                unittest.mock.call._context_factory_factory(
                    base_logger.getChild.return_value,
                    base.metadata,
                    base.certificate_verifier
                ),
                unittest.mock.call.create_starttls_connection(
                    unittest.mock.sentinel.loop,
                    unittest.mock.ANY,
                    host=unittest.mock.sentinel.host,
                    port=unittest.mock.sentinel.port,
                    peer_hostname=unittest.mock.sentinel.host,
                    server_hostname=unittest.mock.sentinel.domain,
                    post_handshake_callback=
                        base.certificate_verifier.post_handshake,
                    ssl_context_factory=unittest.mock.ANY,
                    use_starttls=False,
                ),
            ]
        )

        self.assertEqual(
            result,
            (
                unittest.mock.sentinel.transport,
                base.protocol,
                unittest.mock.sentinel.features,
            )
        )

        timedelta.assert_called_once_with(
            seconds=unittest.mock.sentinel.timeout
        )

        self.assertEqual(
            base.protocol.deadtime_hard_limit,
            timedelta(),
        )

    def test_context_factory(self):
        base = unittest.mock.Mock()

        ssl_context_factory = self.c._context_factory_factory(
            unittest.mock.sentinel.logger,
            base.metadata,
            base.certificate_verifier,
        )

        ssl_context = ssl_context_factory(
            unittest.mock.sentinel.passed_transport)

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.metadata.ssl_context_factory(),
                unittest.mock.call.metadata.
                    ssl_context_factory().set_alpn_protos([b"xmpp-client"]),
                unittest.mock.call.certificate_verifier.setup_context(
                    ssl_context,
                    unittest.mock.sentinel.passed_transport,
                ),
            ]
        )

        self.assertEqual(
            ssl_context,
            base.metadata.ssl_context_factory()
        )

    def test_context_factory_warns_if_set_alpn_protos_is_not_defined(self):
        base_logger = unittest.mock.Mock(spec=logging.Logger)
        base = unittest.mock.Mock()
        del base.metadata.ssl_context_factory.return_value.set_alpn_protos

        context_factory = self.c._context_factory_factory(
            base_logger,
            base.metadata,
            base.certificate_verifier,
        )

        ssl_context = context_factory(unittest.mock.sentinel.passed_transport)

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.metadata.ssl_context_factory(),
                unittest.mock.call.certificate_verifier.setup_context(
                    ssl_context,
                    unittest.mock.sentinel.passed_transport,
                ),
            ]
        )

        self.assertSequenceEqual(
            base_logger.mock_calls,
            [
                unittest.mock.call.warning(
                    "OpenSSL.SSL.Context lacks set_alpn_protos - "
                    "please update pyOpenSSL to a recent version"
                ),
            ]
        )

    def test_context_factory_warns_if_set_alpn_protos_raises(self):
        base_logger = unittest.mock.Mock(spec=logging.Logger)
        base = unittest.mock.Mock()
        base.metadata.ssl_context_factory.return_value.set_alpn_protos.\
            side_effect = NotImplementedError

        context_factory = self.c._context_factory_factory(
            base_logger,
            base.metadata,
            base.certificate_verifier,
        )

        ssl_context = context_factory(unittest.mock.sentinel.passed_transport)

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.metadata.ssl_context_factory(),
                unittest.mock.call.metadata.
                    ssl_context_factory().set_alpn_protos([b"xmpp-client"]),
                unittest.mock.call.certificate_verifier.setup_context(
                    ssl_context,
                    unittest.mock.sentinel.passed_transport,
                ),
            ]
        )

        self.assertSequenceEqual(
            base_logger.mock_calls,
            [
                unittest.mock.call.warning(
                    "the underlying OpenSSL library does not support ALPN"
                ),
            ]
        )

    def test_abort_XMLStream_whin_connect_raises(self):
        captured_features_future = None

        def capture_future(*args, features_future=None, **kwargs):
            nonlocal captured_features_future
            captured_features_future = features_future
            return base.protocol

        features_future = asyncio.Future()
        features_future.set_result(
            unittest.mock.sentinel.features
        )

        base = unittest.mock.Mock()
        base.protocol.starttls = CoroutineMock()
        base.create_starttls_connection = CoroutineMock()
        base.create_starttls_connection.side_effect = Exception()
        base.metadata.tls_required = True
        base.XMLStream.return_value = base.protocol
        base.XMLStream.side_effect = capture_future
        base.Future.return_value = features_future
        base.certificate_verifier.pre_handshake = CoroutineMock()
        base.metadata.certificate_verifier_factory.return_value = \
            base.certificate_verifier
        base.metadata.ssl_context_factory.return_value = \
            unittest.mock.sentinel.ssl_context

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch(
                    "asyncio.Future",
                    new=base.Future,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.ssl_transport.create_starttls_connection",
                    new=base.create_starttls_connection,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.protocol.XMLStream",
                    new=base.XMLStream,
                )
            )

            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.connector.timedelta",
                )
            )

            with self.assertRaises(Exception):
                run_coroutine(self.c.connect(
                    unittest.mock.sentinel.loop,
                    base.metadata,
                    unittest.mock.sentinel.domain,
                    unittest.mock.sentinel.host,
                    unittest.mock.sentinel.port,
                    unittest.mock.sentinel.timeout,
                ))

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.Future(
                    loop=unittest.mock.sentinel.loop,
                ),
                unittest.mock.call.XMLStream(
                    to=unittest.mock.sentinel.domain,
                    features_future=features_future,
                    base_logger=None,
                ),
                unittest.mock.call.metadata.certificate_verifier_factory(),
                unittest.mock.call.certificate_verifier.pre_handshake(
                    unittest.mock.sentinel.domain,
                    unittest.mock.sentinel.host,
                    unittest.mock.sentinel.port,
                    base.metadata,
                ),
                unittest.mock.call.create_starttls_connection(
                    unittest.mock.sentinel.loop,
                    unittest.mock.ANY,
                    host=unittest.mock.sentinel.host,
                    port=unittest.mock.sentinel.port,
                    peer_hostname=unittest.mock.sentinel.host,
                    server_hostname=unittest.mock.sentinel.domain,
                    post_handshake_callback=
                        base.certificate_verifier.post_handshake,
                    ssl_context_factory=unittest.mock.ANY,
                    use_starttls=False,
                ),
                unittest.mock.call.protocol.abort()
            ]
        )



########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import numbers

import aioxmpp.callbacks
import aioxmpp.service
import aioxmpp.structs
import aioxmpp.xso.model


class PresenceClient(aioxmpp.service.Service):
    """
    The presence service tracks all incoming presence information (this does
    not include subscription management stanzas, as these are handled by
    :mod:`aioxmpp.roster`). It is independent of the roster, as directed
    presence is independent of the roster and still needs to be tracked
    accordingly.

    No method to send directed presence is provided; it would basically just
    take a stanza and enqueue it in the clients stream, thus being a mere
    wrapper around :meth:`~.Client.send`, without any benefit.

    The service provides access to presence information summarized by bare JID
    or for each full JID individually. An index over the resources of a bare
    JID is available.

    If an error presence is received for a JID, it replaces all known presence
    information. It is returned for all queries, no matter for which resource
    the query is. As soon as a non-error presence is received for any resource
    of the bare JID, the error is cleared.

    .. automethod:: get_most_available_stanza

    .. automethod:: get_peer_resources

    .. automethod:: get_stanza

    On presence changes of peers, signals are emitted:

    .. signal:: on_bare_available(stanza)

       Fires when the first resource of a peer becomes available.

    .. signal:: on_bare_unavailable(stanza)

       Fires when the last resource of a peer becomes unavailable or enters
       error state.

    .. signal:: on_available(full_jid, stanza)

       Fires when a resource becomes available at `full_jid`. The `stanza`
       which caused the availability is handed over as second argument.

       This signal always fires after :meth:`on_bare_available`.

    .. signal:: on_changed(full_jid, stanza)

       Fires when a the presence of the resource at `full_jid` changes, but
       does not become unavailable or available. The `stanza` which caused the
       change is handed over as second argument.

    .. signal:: on_unavailable(full_jid, stanza)

       Fires when the resource at `full_jid` becomes unavailable, with the
       `stanza` causing the unavailability as second argument.

       This signal always fires before :meth:`on_bare_unavailable`.

       .. note::

          If the resource became unavailable due to an error, the `full_jid`
          will not match the :attr:`~.stanza.StanzaBase.from_` attribute of the
          `stanza`, as the error is coming from the bare JID.

    The three signals :meth:`on_available`,  :meth:`on_changed` and
    :meth:`on_unavailable` never fire for the same stanza.

    .. versionadded:: 0.4

    .. versionchanged:: 0.8

       This class was formerly known as :class:`aioxmpp.presence.Service`. It
       is still available under that name, but the alias will be removed in
       1.0.
    """

    ORDER_AFTER = [
        aioxmpp.dispatcher.SimplePresenceDispatcher,
    ]

    on_bare_available = aioxmpp.callbacks.Signal()
    on_bare_unavailable = aioxmpp.callbacks.Signal()

    on_available = aioxmpp.callbacks.Signal()
    on_changed = aioxmpp.callbacks.Signal()
    on_unavailable = aioxmpp.callbacks.Signal()

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)

        self._presences = {}

    def get_most_available_stanza(self, peer_jid):
        """
        Obtain the stanza describing the most-available presence of the
        contact.

        :param peer_jid: Bare JID of the contact.
        :type peer_jid: :class:`aioxmpp.JID`
        :rtype: :class:`aioxmpp.Presence` or :data:`None`
        :return: The presence stanza of the most available resource or
                 :data:`None` if there is no available resource.

        The "most available" resource is the one whose presence state orderest
        highest according to :class:`~aioxmpp.PresenceState`.

        If there is no available resource for a given `peer_jid`, :data:`None`
        is returned.
        """
        presences = sorted(
            self.get_peer_resources(peer_jid).items(),
            key=lambda item: aioxmpp.structs.PresenceState.from_stanza(item[1])
        )
        if not presences:
            return None
        return presences[-1][1]

    def get_peer_resources(self, peer_jid):
        """
        Return a dict mapping resources of the given bare `peer_jid` to the
        presence state last received for that resource.

        Unavailable presence states are not included. If the bare JID is in a
        error state (i.e. an error presence stanza has been received), the
        returned mapping is empty.
        """
        try:
            d = dict(self._presences[peer_jid])
            d.pop(None, None)
            return d
        except KeyError:
            return {}

    def get_stanza(self, peer_jid):
        """
        Return the last presence recieved for the given bare or full
        `peer_jid`. If the last presence was unavailable, the return value is
        :data:`None`, as if no presence was ever received.

        If no presence was ever received for the given bare JID, :data:`None`
        is returned.
        """
        try:
            return self._presences[peer_jid.bare()][peer_jid.resource]
        except KeyError:
            pass
        try:
            return self._presences[peer_jid.bare()][None]
        except KeyError:
            pass

    @aioxmpp.dispatcher.presence_handler(
        aioxmpp.structs.PresenceType.AVAILABLE,
        None)
    @aioxmpp.dispatcher.presence_handler(
        aioxmpp.structs.PresenceType.UNAVAILABLE,
        None)
    @aioxmpp.dispatcher.presence_handler(
        aioxmpp.structs.PresenceType.ERROR,
        None)
    def handle_presence(self, st):
        bare = st.from_.bare()
        resource = st.from_.resource

        if st.type_ == aioxmpp.structs.PresenceType.UNAVAILABLE:
            try:
                dest_dict = self._presences[bare]
            except KeyError:
                return
            dest_dict.pop(None, None)
            if resource in dest_dict:
                self.on_unavailable(st.from_, st)
                if len(dest_dict) == 1:
                    self.on_bare_unavailable(st)
                del dest_dict[resource]
        elif st.type_ == aioxmpp.structs.PresenceType.ERROR:
            try:
                dest_dict = self._presences[bare]
            except KeyError:
                pass
            else:
                for resource in dest_dict.keys():
                    self.on_unavailable(st.from_.replace(resource=resource),
                                        st)
                self.on_bare_unavailable(st)
            self._presences[bare] = {None: st}
        else:
            dest_dict = self._presences.setdefault(bare, {})
            dest_dict.pop(None, None)
            bare_became_available = not dest_dict
            resource_became_available = resource not in dest_dict
            dest_dict[resource] = st

            if bare_became_available:
                self.on_bare_available(st)
            if resource_became_available:
                self.on_available(st.from_, st)
            else:
                self.on_changed(st.from_, st)


class PresenceServer(aioxmpp.service.Service):
    """
    Manage the presence broadcast by the client.

    .. .. note::

    ..    This was formerly handled by the :class:`aioxmpp.PresenceManagedClient`,
    ..    which is now merely a shim wrapper around :class:`aioxmpp.Client` and
    ..    :class:`PresenceServer`.

    The :class:`PresenceServer` manages broadcasting and re-broadcasting the
    presence of the client as needed.

    The presence state is initialised to an unavailable presence. Unavailable
    presences are not emitted when the stream is established.

    Presence information:

    .. autoattribute:: state

    .. autoattribute:: status

    .. autoattribute:: priority

    .. automethod:: make_stanza

    Changing/sending/watching presence:

    .. automethod:: set_presence

    .. automethod:: resend_presence

    .. signal:: on_presence_changed()

       Emits after the presence has been changed in the
       :class:`PresenceServer`.

    .. signal:: on_presence_state_changed(new_state)

       Emits after the presence *state* has been changed in the
       :class:`PresenceServer`.

       This signal does not emit if other parts of the presence (such as
       priority or status texts) change, while the presence state itself stays
       the same.

    .. versionadded:: 0.8
    """

    on_presence_changed = aioxmpp.callbacks.Signal()
    on_presence_state_changed = aioxmpp.callbacks.Signal()

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)
        self._state = aioxmpp.PresenceState(False)
        self._status = {}
        self._priority = 0

        client.before_stream_established.connect(
            self._before_stream_established
        )

    @asyncio.coroutine
    def _before_stream_established(self):
        if not self._state.available:
            return True

        yield from self.client.send(self.make_stanza())

        return True

    @property
    def state(self):
        """
        The currently set presence state (as :class:`aioxmpp.PresenceState`)
        which is broadcast when the client connects and when the presence is
        re-emitted.

        This attribute cannot be written. It does not reflect the actual
        presence seen by others. For example when the client is in fact
        offline, others will see unavailable presence no matter what is set
        here.
        """
        return self._state

    @property
    def status(self):
        """
        The currently set textual presence status which is broadcast when the
        client connects and when the presence is re-emitted.

        This attribute cannot be written. It does not reflect the actual
        presence seen by others. For example when the client is in fact
        offline, others will see unavailable presence no matter what is set
        here.
        """
        return self._status

    @property
    def priority(self):
        """
        The currently set priority which is broadcast when the client connects
        and when the presence is re-emitted.

        This attribute cannot be written. It does not reflect the actual
        presence seen by others. For example when the client is in fact
        offline, others will see unavailable presence no matter what is set
        here.
        """
        return self._priority

    def make_stanza(self):
        """
        Create and return a presence stanza with the current settings.

        :return: Presence stanza
        :rtype: :class:`aioxmpp.Presence`
        """
        stanza = aioxmpp.Presence()
        self._state.apply_to_stanza(stanza)
        stanza.status.update(self._status)
        return stanza

    def set_presence(self, state, status={}, priority=0):
        """
        Change the presence broadcast by the client.

        :param state: New presence state to broadcast
        :type state: :class:`aioxmpp.PresenceState`
        :param status: New status information to broadcast
        :type status: :class:`dict` or :class:`str`
        :param priority: New priority for the resource
        :type priority: :class:`int`
        :return: Stanza token of the presence stanza or :data:`None` if the
                 presence is unchanged or the stream is not connected.
        :rtype: :class:`~.stream.StanzaToken`

        If the client is currently connected, the new presence is broadcast
        immediately.

        `status` must be either a string or something which can be passed to
        the :class:`dict` constructor. If it is a string, it is wrapped into a
        dict using ``{None: status}``. The mapping must map
        :class:`~.LanguageTag` objects (or :data:`None`) to strings. The
        information will be used to generate internationalised presence status
        information. If you do not need internationalisation, simply use the
        string version of the argument.
        """

        if not isinstance(priority, numbers.Integral):
            raise TypeError(
                "invalid priority: got {}, expected integer".format(
                    type(priority)
                )
            )

        if not isinstance(state, aioxmpp.PresenceState):
            raise TypeError(
                "invalid state: got {}, expected aioxmpp.PresenceState".format(
                    type(state),
                )
            )

        if isinstance(status, str):
            new_status = {None: status}
        else:
            new_status = dict(status)
        new_priority = int(priority)

        emit_state_event = self._state != state
        emit_overall_event = (
            emit_state_event or
            self._priority != new_priority or
            self._status != new_status
        )

        self._state = state
        self._status = new_status
        self._priority = new_priority

        if emit_state_event:
            self.on_presence_state_changed()
        if emit_overall_event:
            self.on_presence_changed()
            return self.resend_presence()

    def resend_presence(self):
        """
        Re-send the currently configured presence.

        :return: Stanza token of the presence stanza or :data:`None` if the
                 stream is not established.
        :rtype: :class:`~.stream.StanzaToken`

        .. note::

           :meth:`set_presence` automatically broadcasts the new presence if
           any of the parameters changed.
        """

        if self.client.established:
            return self.client.enqueue(self.make_stanza())



#!/usr/bin/env python
########################################################################
# File name: travis-e2etest-metronome.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
from __future__ import unicode_literals, print_function

import subprocess
import os
import time
import sys

metronome_version = os.environ["METRONOME_VERSION"]

metronome = subprocess.Popen(
    [
        "./metronome",
    ],
    cwd=os.path.join(os.getcwd(), "metronome"),
)

time.sleep(2)

if metronome.poll() is not None:
    print("Metronome already terminated!", file=sys.stderr)
    sys.exit(10)

try:
    subprocess.check_call(
        [
            "python3",
            "-m", "aioxmpp.e2etest",
            "--e2etest-config=utils/metronome-cfg/{}/e2etest.ini".format(
                metronome_version,
            ),
            "tests"
        ]
    )
finally:
    metronome.kill()
    metronome.wait()



########################################################################
# File name: delay.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces

from ..stanza import Message


namespaces.xep0203_delay = "urn:xmpp:delay"


class Delay(xso.XSO):
    """
    A marker indicating delayed delivery of a stanza.

    .. attribute:: from_

       The address as :class:`aioxmpp.JID` of the entity where the stanza was
       delayed. May be :data:`None`.

    .. attribute:: stamp

       The timestamp (as :class:`datetime.datetime`) at which the stanza was
       originally sent or intended to be sent.

    .. attribute:: reason

       The reason for which the stanza was delayed or :data:`None`.

    .. warning::

       Please take the security considerations of :xep:`203` into account.

    """

    TAG = namespaces.xep0203_delay, "delay"

    from_ = xso.Attr(
        "from",
        type_=xso.JID(),
        default=None,
    )

    stamp = xso.Attr(
        "stamp",
        type_=xso.DateTime(),
    )

    reason = xso.Text(
        default=None
    )


Message.xep0203_delay = xso.ChildList([Delay])



########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp
import aioxmpp.xso as xso
import aioxmpp.service as service

from aioxmpp.utils import namespaces

from . import xso as vcard_xso


class VCardService(service.Service):
    """
    Service for handling vcard-temp.

    .. automethod:: get_vcard

    .. automethod:: set_vcard
    """

    @asyncio.coroutine
    def get_vcard(self, jid=None):
        """
        Get the vCard stored for the jid `jid`. If `jid` is
        :data:`None` get the vCard of the connected entity.

        :param jid: the object to retrieve.
        :returns: the stored vCard.

        We mask a :class:`XMPPCancelError` in case it is
        ``feature-not-implemented`` or ``item-not-found`` and return
        an empty vCard, since this can be understood to be semantically
        equivalent.
        """

        iq = aioxmpp.IQ(
            type_=aioxmpp.IQType.GET,
            to=jid,
            payload=vcard_xso.VCard(),
        )

        try:
            return (yield from self.client.send(iq))
        except aioxmpp.XMPPCancelError as e:
            if e.condition in (
                    aioxmpp.ErrorCondition.FEATURE_NOT_IMPLEMENTED,
                    aioxmpp.ErrorCondition.ITEM_NOT_FOUND):
                return vcard_xso.VCard()
            else:
                raise

    @asyncio.coroutine
    def set_vcard(self, vcard):
        """
        Store the vCard `vcard` for the connected entity.

        :param vcard: the vCard to store.

        .. note::

           `vcard` should always be derived from the result of
           `get_vcard` to preserve the elements of the vcard the
           client does not modify.

        .. warning::

           It is in the responsibility of the user to supply valid
           vcard data as per :xep:`0054`.
        """
        iq = aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            payload=vcard,
        )
        yield from self.client.send(iq)



########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest
import unittest.mock
import aioxmpp
import asyncio

import aioxmpp.ibr.service as ibr_service
import aioxmpp.ibr.xso as ibr_xso

from aioxmpp.testutils import (
    make_connected_client,
    run_coroutine,
    CoroutineMock
)


TEST_PEER = aioxmpp.JID.fromstr("juliet@capulet.lit/balcony")


class TestService(unittest.TestCase):

    def setUp(self):
        self.cc = make_connected_client()
        self.s = ibr_service.RegistrationService(self.cc)

    def test_ibr_get_client_info(self):
        query = ibr_xso.Query()
        query.username = TEST_PEER.localpart
        self.cc.send.return_value = query
        self.cc.local_jid = TEST_PEER

        res = run_coroutine(self.s.get_client_info())

        self.cc.send.assert_called_once_with(unittest.mock.ANY)

        _, (iq, ), _ = self.cc.send.mock_calls[0]

        self.assertIsInstance(
            iq,
            aioxmpp.IQ,
        )

        self.assertEqual(
            iq.to,
            TEST_PEER.bare().replace(localpart=None),
        )

        self.assertEqual(
            iq.type_,
            aioxmpp.IQType.GET,
        )

        self.assertIsInstance(
            iq.payload,
            ibr_xso.Query,
        )

        self.assertIsInstance(
            res,
            ibr_xso.Query,
        )

        self.assertEqual(
            res.username,
            TEST_PEER.localpart,
        )

    def test_ibr_change_pass(self):
        self.cc.local_jid = TEST_PEER
        new_pass = "aaa"

        run_coroutine(
            self.s.change_pass(new_pass)
        )

        self.cc.send.assert_called_once_with(unittest.mock.ANY)

        _, (iq, ), _ = self.cc.send.mock_calls[0]

        self.assertIsInstance(
            iq,
            aioxmpp.IQ,
        )

        self.assertEqual(
            iq.type_,
            aioxmpp.IQType.SET,
        )

        self.assertEqual(
            iq.to,
            TEST_PEER.bare().replace(localpart=None),
        )

        self.assertIsInstance(
            iq.payload,
            ibr_xso.Query,
        )

        self.assertEqual(
            iq.payload.username,
            TEST_PEER.localpart,
        )

        self.assertEqual(
            iq.payload.password,
            new_pass,
        )

    def test_ibr_cancel_registration(self):
        self.cc.local_jid = TEST_PEER

        run_coroutine(self.s.cancel_registration())

        self.cc.send.assert_called_once_with(unittest.mock.ANY)

        _, (iq, ), _ = self.cc.send.mock_calls[0]

        self.assertIsInstance(
            iq,
            aioxmpp.IQ,
        )

        self.assertEqual(
            iq.to,
            TEST_PEER.bare().replace(localpart=None),
        )

        self.assertEqual(
            iq.type_,
            aioxmpp.IQType.SET,
        )

        self.assertIsInstance(
            iq.payload,
            ibr_xso.Query,
        )

        self.assertEqual(
            iq.payload.remove,
            True,
        )

    def test_ibr_get_registration_fields(self):
        with unittest.mock.patch(
                'aioxmpp.protocol.send_and_wait_for',
                new=CoroutineMock()
        ) as mock1:
            iq_res = aioxmpp.IQ(
                type_=aioxmpp.IQType.GET,
                payload=ibr_xso.Query()
            )
            iq_res.payload.username = ''
            mock1.return_value = iq_res
            stream = unittest.mock.Mock(spec=aioxmpp.protocol.XMLStream)
            stream._to = TEST_PEER.domain
            res = run_coroutine(aioxmpp.ibr.get_registration_fields(stream))

            _, (_, iq, *_), _ = mock1.mock_calls[0]

            iq = iq[0]

            self.assertIsInstance(
                iq,
                aioxmpp.IQ,
            )

            self.assertIs(
                iq_res.payload,
                res,
            )

            self.assertIsInstance(
                res,
                ibr_xso.Query,
            )

            self.assertEqual(
                res.username,
                '',
            )

    def test_ibr_register(self):
        password = "aaa"
        aux_fields = {"nick": "romeo's lover"}
        with unittest.mock.patch(
                'aioxmpp.protocol.send_and_wait_for',
                new=CoroutineMock()
        ) as mock1:
            stream = unittest.mock.Mock(spec=aioxmpp.protocol.XMLStream)
            stream._to = TEST_PEER.domain
            query = aioxmpp.ibr.Query(TEST_PEER.localpart, "aaa", aux_fields)
            run_coroutine(aioxmpp.ibr.register(stream, query))

            _, (_, iq, *_), _ = mock1.mock_calls[0]

            iq = iq[0]

            self.assertIsInstance(
                iq,
                aioxmpp.IQ,
            )

            self.assertEqual(
                iq.to,
                TEST_PEER.bare().replace(localpart=None),
            )

            self.assertEqual(
                iq.type_,
                aioxmpp.IQType.SET,
            )

            self.assertIsInstance(
                iq.payload,
                ibr_xso.Query,
            )

            self.assertEqual(
                iq.payload.username,
                TEST_PEER.localpart,
            )

            self.assertEqual(
                iq.payload.password,
                password,
            )

            if aux_fields is not None:
                for key, value in aux_fields.items():
                    self.assertEqual(
                        getattr(iq.payload, key),
                        value,
                    )


if __name__ == '__main__':
    unittest.main()
